################################################################################
# Copyright (C) 2013-2014 Jaakko Luttinen
#
# This file is licensed under the MIT License.
################################################################################


import warnings

import numpy as np

from bayespy.utils import misc

from .node import ensureparents
from .stochastic import Stochastic, Distribution

class ExponentialFamilyDistribution(Distribution):
    """
    Sub-classes implement distribution specific computations.
    """

    #
    # The following methods are for ExponentialFamily distributions
    #

    def compute_message_to_parent(self, parent, index, u_self, *u_parents):
        raise NotImplementedError()

    def compute_phi_from_parents(self, *u_parents, mask=True):
        raise NotImplementedError()

    def compute_moments_and_cgf(self, phi, mask=True):
        raise NotImplementedError()

    #
    # The following methods are for Mixture class
    #

    def compute_cgf_from_parents(self, *u_parents):
        raise NotImplementedError()

    def compute_fixed_moments_and_f(self, x, mask=True):
        raise NotImplementedError()

    def compute_logpdf(self, u, phi, g, f, ndims):
        """ Compute E[log p(X)] given E[u], E[phi], E[g] and
        E[f]. Does not sum over plates."""

        # TODO/FIXME: Should I take into account what is latent or
        # observed, or what is even totally ignored (by the mask).
        L = g + f
        for (phi_i, u_i, ndims_i) in zip(phi, u, ndims):
            # Axes to sum (dimensions of the variable, not the plates)
            axis_sum = tuple(range(-ndims_i,0))
            # Compute the term
            # TODO/FIXME: Use einsum!
            L = L + np.sum(
                np.where(u_i != 0, phi_i, 0) * u_i,
                axis=axis_sum
            )
        return L


    def compute_gradient(self, g, u, phi):
        r"""
        Compute the standard gradient with respect to the natural parameters.
        """

        raise NotImplementedError("Standard gradient not yet implemented for %s"
                                  % (self.__class__.__name__))



def useconstructor(__init__):
    def constructor_decorator(self, *args, **kwargs):
        if (self.dims is None or
            self._distribution is None or
            self._moments is None or
            self._parent_moments is None):

            (args, kwargs, dims, plates, dist, stats, pstats) = \
              self._constructor(*args, **kwargs)

            self.dims = dims
            self._distribution = dist
            self._moments = stats
            self._parent_moments = pstats
            self.plates = plates

        __init__(self, *args, **kwargs)

    return constructor_decorator

class ExponentialFamily(Stochastic):
    """
    A base class for nodes using natural parameterization `phi`.

    phi

    Sub-classes must implement the following static methods:
       _compute_message_to_parent(index, u_self, *u_parents)
       _compute_phi_from_parents(*u_parents, mask)
       _compute_moments_and_cgf(phi, mask)
       _compute_fixed_moments_and_f(x, mask=True)

    Sub-classes may need to re-implement:
    1. If they manipulate plates:
       _compute_weights_to_parent(index, weights)
       _compute_plates_to_parent(self, index, plates)
       _compute_plates_from_parent(self, index, plates)

    """

    # Sub-classes should overwrite this (possibly using _constructor)
    dims = None

    # Sub-classes should overwrite this
    _distribution = None

    @useconstructor
    def __init__(self, *parents, initialize=True, phi_bias=None, **kwargs):

        self.annealing = 1.0

        # Terms for the lower bound (G for latent and F for observed)
        self.g = np.array(np.nan)
        self.f = np.array(np.nan)

        self._phi_bias = phi_bias if phi_bias is not None else len(self.dims) * [0.0]

        super().__init__(*parents,
                         initialize=initialize,
                         dims=self.dims,
                         **kwargs)

        if not initialize:
            axes = len(self.plates)*(1,)
            self.phi = [misc.nans(axes+dim) for dim in self.dims]


    @classmethod
    @ensureparents
    def _constructor(cls, *parents, **kwargs):
        """
        Constructs distribution and moments objects.

        If __init__ uses useconstructor decorator, this method is called to
        construct distribution and moments objects.

        The method is given the same inputs as __init__. For some nodes, some of
        these can't be "static" class attributes, then the node class must
        overwrite this method to construct the objects manually.

        The point of distribution class is to move general distribution but
        not-node specific code. The point of moments class is to define the
        messaging protocols.
        """
        parent_plates = [cls._distribution.plates_from_parent(ind, parent.plates)
                         for (ind, parent) in enumerate(parents)]
        return (parents,
                kwargs,
                cls.dims,
                cls._total_plates(kwargs.get('plates'), *parent_plates),
                cls._distribution,
                cls._moments,
                cls._parent_moments)

    def _initialize_from_parent_moments(self, *u_parents):
        if not np.all(self.observed):
            # Update natural parameters using parents
            self._update_phi_from_parents(*u_parents)

            # Update moments
            mask = np.logical_not(self.observed)
            (u, g) = self._distribution.compute_moments_and_cgf(self.phi,
                                                                mask=mask)
            # TODO/FIXME/BUG: You should use observation mask in order to not
            # overwrite them!
            self._set_moments_and_cgf(u, g, mask=mask)


    def initialize_from_prior(self):
        u_parents = self._message_from_parents()
        self._initialize_from_parent_moments(*u_parents)


    def initialize_from_parameters(self, *args):
        u_parents = [p_mom.compute_fixed_moments(x)
                     for (p_mom, x) in zip(self._parent_moments, args)]
        self._initialize_from_parent_moments(*u_parents)


    def initialize_from_value(self, x, *args):
        # Update moments from value
        mask = np.logical_not(self.observed)
        u = self._moments.compute_fixed_moments(x, *args)
        # Check that the shape is correct
        for i in range(len(u)):
            ndim = len(self.dims[i])
            if ndim > 0:
                if np.shape(u[i])[-ndim:] != self.dims[i]:
                    raise ValueError("The initial value for node %s has invalid shape %s."
                                     % (np.shape(x)))
        self._set_moments_and_cgf(u, np.inf, mask=mask)

    def initialize_from_random(self):
        """
        Set the variable to a random sample from the current distribution.
        """
        #self.initialize_from_prior()
        X = self.random()
        self.initialize_from_value(X)


    def _update_phi_from_parents(self, *u_parents):

        # TODO/FIXME: Could this be combined to the function
        # _update_distribution_and_lowerbound ?
        # No, because some initialization methods may want to use this.

        # This makes correct broadcasting
        self.phi = [
            a + b
            for (a, b) in zip(
                self._distribution.compute_phi_from_parents(*u_parents),
                self._phi_bias
            )
        ]
        # Make sure phi has the correct number of axes. It makes life
        # a bit easier elsewhere.
        for i in range(len(self.phi)):
            axes = len(self.plates) + self.ndims[i] - np.ndim(self.phi[i])
            if axes > 0:
                # Add axes
                self.phi[i] = misc.add_leading_axes(self.phi[i], axes)
            elif axes < 0:
                # Remove extra leading axes
                first = -(len(self.plates)+self.ndims[i])
                sh = np.shape(self.phi[i])[first:]
                self.phi[i] = np.reshape(self.phi[i], sh)
            # Check that the shape is correct
            if not misc.is_shape_subset(np.shape(self.phi[i]),
                                         self.get_shape(i)):
                raise ValueError("Incorrect shape of phi[%d] in node class %s. "
                                 "Shape is %s but it should be broadcastable "
                                 "to shape %s."
                                 % (i,
                                    self.__class__.__name__,
                                    np.shape(self.phi[i]),
                                    self.get_shape(i)))

    def _set_moments_and_cgf(self, u, g, mask=True):
        self._set_moments(u, mask=mask)

        self.g = np.where(mask, g, self.g)

        return


    def get_riemannian_gradient(self):
        r"""
        Computes the Riemannian/natural gradient.
        """
        u_parents = self._message_from_parents()
        m_children = self._message_from_children()

        # TODO/FIXME: Put observed plates to zero?
        # Compute the gradient
        phi = [
            a + b
            for (a, b) in zip(
                self._distribution.compute_phi_from_parents(*u_parents),
                self._phi_bias
            )
        ]
        for i in range(len(self.phi)):
            phi[i] = self.annealing * (phi[i] + m_children[i]) - self.phi[i]
            phi[i] = phi[i] * np.ones(self.get_shape(i))

        return phi


    def get_gradient(self, rg):
        r""" Computes gradient with respect to the natural parameters.

        The function takes the Riemannian gradient as an input.  This is for
        three reasons: 1) You probably want to use the Riemannian gradient
        anyway so this helps avoiding accidental use of this function.  2) The
        gradient is computed by using the Riemannian gradient and chain rules.
        3) Probably you need both Riemannian and normal gradients anyway so you
        can provide it to this function to avoid re-computing it."""

        g = self._distribution.compute_gradient(rg, self.u, self.phi)
        for i in range(len(g)):
            g[i] /= self.annealing
        return g


    ## def update_parameters(self, d, scale=1.0):
    ##     r"""
    ##     Update the parameters of the VB distribution given a change.

    ##     The parameters should be such that they can be used for
    ##     optimization, that is, use log transformation for positive
    ##     parameters.
    ##     """
    ##     phi = self.get_parameters()
    ##     for i in range(len(phi)):
    ##         phi[i] = phi[i] + scale*d[i]
    ##     self.set_parameters(phi)
    ##     return


    def get_parameters(self):
        r"""
        Return parameters of the VB distribution.

        The parameters should be such that they can be used for
        optimization, that is, use log transformation for positive
        parameters.
        """
        return [np.copy(p) for p in self.phi]



    def _decode_parameters(self, x):
        return [np.copy(p) for p in x]


    def set_parameters(self, x):
        r"""
        Set the parameters of the VB distribution.

        The parameters should be such that they can be used for
        optimization, that is, use log transformation for positive
        parameters.
        """
        self.phi = self._decode_parameters(x)
        self._update_moments_and_cgf()
        return


    def _update_distribution_and_lowerbound(self, m_children, *u_parents):

        # Update phi first from parents..
        self._update_phi_from_parents(*u_parents)
        # .. then just add children's message
        self.phi = [self.annealing * (phi + m)
                    for (phi, m) in zip(self.phi, m_children)]

        # Update u and g
        self._update_moments_and_cgf()


    def _update_moments_and_cgf(self):
        """
        Update moments and cgf based on current phi.
        """
        # Mask for plates to update (i.e., unobserved plates)
        update_mask = np.logical_not(self.observed)

        # Compute the moments (u) and CGF (g)...
        (u, g) = self._distribution.compute_moments_and_cgf(self.phi,
                                                            mask=update_mask)
        # ... and store them
        self._set_moments_and_cgf(u, g, mask=update_mask)


    def observe(self, x, *args, mask=True):
        """
        Fix moments, compute f and propagate mask.
        """

        # Compute fixed moments
        (u, f) = self._distribution.compute_fixed_moments_and_f(x, *args,
                                                                mask=mask)

        # # Check the dimensionality of the observations
        # self._check_shape()
        # for (i,v) in enumerate(u):
        #     # This is what the dimensionality "should" be
        #     s = self.plates + self.dims[i]
        #     t = np.shape(v)
        #     if s != t:
        #         msg = "Dimensionality of the observations incorrect."
        #         msg += "\nShape of input: " + str(t)
        #         msg += "\nExpected shape: " + str(s)
        #         msg += "\nCheck plates."
        #         raise Exception(msg)

        # Set the moments. Shape checking is done there.
        self._set_moments(u, mask=mask, broadcast=False)

        self.f = np.where(mask, f, self.f)

        # Observed nodes should not be ignored
        self.observed = mask
        self._update_mask()

    def lower_bound_contribution(self, gradient=False, ignore_masked=True):
        r"""Compute E[ log p(X|parents) - log q(X) ]

        If deterministic annealing is used, the term E[ -log q(X) ] is
        divided by the anneling coefficient.  That is, phi and cgf of q
        are multiplied by the temperature (inverse annealing
        coefficient).

        """

        # Annealing temperature
        T = 1 / self.annealing

        # Messages from parents
        u_parents = self._message_from_parents()
        phi = [
            a # + b # TODO: Should the bias be added here or not?
            for (a, b) in zip(
                self._distribution.compute_phi_from_parents(*u_parents),
                self._phi_bias
            )
        ]
        # G from parents
        L = self._distribution.compute_cgf_from_parents(*u_parents)

        # G for unobserved variables (ignored variables are handled properly
        # automatically)
        latent_mask = np.logical_not(self.observed)

        # G and F
        if np.all(self.observed):
            z = np.nan
        elif T == 1:
            z = -self.g
        else:
            z = -T * self.g
            ## TRIED THIS BUT IT WAS WRONG:
            ## z = -T * self.g + (1-T) * self.f
            ## if np.any(np.isnan(self.f)):
            ##     warnings.warn("F(x) not implemented for node %s. This "
            ##                   "is required for annealed lower bound "
            ##                   "computation." % self.__class__.__name__)
            ##
            ## It was wrong because the optimal q distribution has f which is
            ## weighted by 1/T and here the f of q is weighted by T so the
            ## total weight is 1, thus it cancels out with f of p.

        L = L + np.where(self.observed, self.f, z)

        for (phi_p, phi_q, u_q, dims) in zip(phi, self.phi, self.u, self.dims):
            # Form a mask which puts observed variables to zero and
            # broadcasts properly
            latent_mask_i = misc.add_trailing_axes(
                                misc.add_leading_axes(
                                    latent_mask,
                                    len(self.plates) - np.ndim(latent_mask)),
                                len(dims))
            axis_sum = tuple(range(-len(dims),0))

            # Compute the term
            phi_q = np.where(latent_mask_i, phi_q, 0)
            # Apply annealing
            phi_diff = phi_p - T * phi_q
            # Handle 0 * -inf
            phi_diff = np.where(u_q != 0, phi_diff, 0)
            # TODO/FIXME: Use einsum here?
            Z = np.sum(phi_diff * u_q, axis=axis_sum)

            L = L + Z

        if ignore_masked:
            return (np.sum(np.where(self.mask, L, 0))
                    * self.broadcasting_multiplier(self.plates,
                                                   np.shape(L),
                                                   np.shape(self.mask))
                    * np.prod(self.plates_multiplier))
        else:
            return (np.sum(L)
                    * self.broadcasting_multiplier(self.plates,
                                                   np.shape(L))
                    * np.prod(self.plates_multiplier))


    def logpdf(self, X, mask=True):
        """
        Compute the log probability density function Q(X) of this node.
        """
        if mask is not True:
            raise NotImplementedError('Mask not yet implemented')
        (u, f) = self._distribution.compute_fixed_moments_and_f(X, mask=mask)
        Z = 0
        for (phi_d, u_d, dims) in zip(self.phi, u, self.dims):
            axis_sum = tuple(range(-len(dims),0))
            # TODO/FIXME: Use einsum here?
            Z = Z + np.sum(phi_d * u_d, axis=axis_sum)
            #Z = Z + misc.sum_multiply(phi_d, u_d, axis=axis_sum)

        return (self.g + f + Z)


    def pdf(self, X, mask=True):
        """
        Compute the probability density function of this node.
        """
        return np.exp(self.logpdf(X, mask=mask))


    def _save(self, group):
        """
        Save the state of the node into a HDF5 file.

        group can be the root
        """
        ## if name is None:
        ##     name = self.name
        ## subgroup = group.create_group(name)

        for i in range(len(self.phi)):
            misc.write_to_hdf5(group, self.phi[i], 'phi%d' % i)
        misc.write_to_hdf5(group, self.f, 'f')
        misc.write_to_hdf5(group, self.g, 'g')
        super()._save(group)


    def _load(self, group):
        """
        Load the state of the node from a HDF5 file.
        """
        # TODO/FIXME: Check that the shapes are correct!
        for i in range(len(self.phi)):
            phii = group['phi%d' % i][...]
            self.phi[i] = phii

        self.f = group['f'][...]
        self.g = group['g'][...]
        super()._load(group)


    def random(self):
        """
        Draw a random sample from the distribution.
        """
        return self._distribution.random(*(self.phi), plates=self.plates)



################################################################################
# Copyright (C) 2013-2014 Jaakko Luttinen
#
# This file is licensed under the MIT License.
################################################################################


"""
Unit tests for `dot` module.
"""

import unittest


import numpy as np
import scipy

from numpy import testing

from ..node import Node, Moments

from ...vmp import VB

from bayespy.utils import misc


class TestMoments(unittest.TestCase):

    def test_converter(self):
        """
        Tests complex conversions for moment classes
        """
        
        # Simple one step conversions
        class A(Moments):
            pass
        class B(Moments):
            _converters = {A: lambda x: x+1}
        f = B().get_converter(B)
        self.assertEqual(f(3), 3)
        f = B().get_converter(Moments)
        self.assertEqual(f(3), 3)
        f = B().get_converter(A)
        self.assertEqual(f(3), 4)
        f = A().get_converter(A)
        self.assertEqual(f(3), 3)
        f = A().get_converter(Moments)
        self.assertEqual(f(3), 3)
        self.assertRaises(Moments.NoConverterError,
                          A().get_converter,
                          B)

        # Convert via parent
        class C(B):
            pass
        f = C().get_converter(A)
        self.assertEqual(f(3), 4)

        # Convert via grand parent
        class D(C):
            pass
        class E(D):
            pass
        f = E().get_converter(A)
        self.assertEqual(f(3), 4)

        # Can't convert to child
        self.assertRaises(Moments.NoConverterError,
                          Moments().get_converter,
                          A)
        # Convert to grand child
        class F(Moments):
            _converters = {E: lambda x: 2*x}
        f = F().get_converter(B)
        self.assertEqual(f(3), 6)

        # Use two conversions
        f = F().get_converter(A)
        self.assertEqual(f(3), 2*3+1)
        
        # Can't use child's converter
        class H(Moments):
            pass
        class I(Moments):
            _converters = {A: lambda x: x+1}
        self.assertRaises(Moments.NoConverterError,
                          H().get_converter,
                          A)

        # Conversion to parent is not success
        class J(A):
            pass
        self.assertRaises(Moments.NoConverterError,
                          I().get_converter,
                          J)
        
        # Infinite loop
        class X(Moments):
            pass
        class Y(Moments):
            pass
        X.add_converter(Y, lambda x: x+1)
        Y.add_converter(X, lambda x: x+1)
        self.assertRaises(Moments.NoConverterError,
                          X().get_converter,
                          A)

        # Test that add_converter function does not change the converters of
        # parent classes
        class Z(Moments):
            pass
        class W(Z):
            pass
        W.add_converter(Y, lambda x: x)
        self.assertRaises(Moments.NoConverterError,
                          Z().get_converter,
                          Y)

        # Test that after using add_converter for a child class and then for the
        # parent class, the child class is still able to use the parent's
        # converters
        class X(Moments):
            pass
        class Y(Moments):
            pass
        class A(Moments):
            pass
        class B(A):
            pass
        B.add_converter(Y, lambda x: x+1)
        A.add_converter(X, lambda x: 2*x)
        f = B().get_converter(X)
        self.assertEqual(f(3), 6)

        pass
        
    
class TestNode(misc.TestCase):

    def check_message_to_parent(self, plates_child, plates_message,
                                plates_mask, plates_parent, dims=(2,)):

        # Dummy message
        msg = np.random.randn(*(plates_message+dims))
        # Mask with every other True and every other False
        mask = np.mod(np.arange(np.prod(plates_mask)).reshape(plates_mask),
                      2) == 0

        # Set up the dummy model
        class Dummy(Node):
            _moments = Moments()
            def __init__(self, *args, **kwargs):
                self._parent_moments = len(args)*(Moments(),)
                super().__init__(*args, **kwargs)
            def _get_message_and_mask_to_parent(self, index, u_parent=None):
                return ([msg], mask)
            def _get_id_list(self):
                return []
        parent = Dummy(dims=[dims], plates=plates_parent)
        child = Dummy(parent, dims=[dims], plates=plates_child)

        m = child._message_to_parent(0)[0] * np.ones(plates_parent+dims)

        # Brute-force computation of the message without too much checking
        m_true = msg * misc.squeeze(mask[...,np.newaxis]) * np.ones(plates_child+dims)
        for ind in range(len(plates_child)):
            axis = -ind - 2
            if ind >= len(plates_parent):
                m_true = np.sum(m_true, axis=axis, keepdims=False)
            elif plates_parent[-ind-1] == 1:
                m_true = np.sum(m_true, axis=axis, keepdims=True)

        testing.assert_allclose(m, m_true,
                                err_msg="Incorrect message.")

    def test_message_to_parent(self):
        """
        Test plate handling in _message_to_parent.
        """

        # Test empty plates with scalar messages
        self.check_message_to_parent((),
                                     (),
                                     (),
                                     (),
                                     dims=())
        # Test singular plates
        self.check_message_to_parent((2,3,4),
                                     (2,3,4),
                                     (2,3,4),
                                     (2,3,4))
        self.check_message_to_parent((2,3,4),
                                     (2,1,4),
                                     (2,3,4),
                                     (2,3,4))
        self.check_message_to_parent((2,3,4),
                                     (2,3,4),
                                     (2,1,4),
                                     (2,3,4))
        self.check_message_to_parent((2,3,4),
                                     (2,3,4),
                                     (2,3,4),
                                     (2,1,4))
        self.check_message_to_parent((2,3,4),
                                     (2,1,4),
                                     (2,1,4),
                                     (2,3,4))
        self.check_message_to_parent((2,3,4),
                                     (2,3,4),
                                     (2,1,4),
                                     (2,1,4))
        self.check_message_to_parent((2,3,4),
                                     (2,1,4),
                                     (2,3,4),
                                     (2,1,4))
        self.check_message_to_parent((2,3,4),
                                     (2,1,4),
                                     (2,1,4),
                                     (2,1,4))
        # Test missing plates
        self.check_message_to_parent((4,3),
                                     (4,3),
                                     (4,3),
                                     (4,3))
        self.check_message_to_parent((4,3),
                                     (  3,),
                                     (4,3),
                                     (4,3))
        self.check_message_to_parent((4,3),
                                     (4,3),
                                     (  3,),
                                     (4,3))
        self.check_message_to_parent((4,3),
                                     (4,3),
                                     (4,3),
                                     (  3,))
        self.check_message_to_parent((4,3),
                                     (  3,),
                                     (  3,),
                                     (4,3))
        self.check_message_to_parent((4,3),
                                     (  3,),
                                     (4,3),
                                     (  3,))
        self.check_message_to_parent((4,3),
                                     (4,3),
                                     (  3,),
                                     (  3,))
        self.check_message_to_parent((4,3),
                                     (  3,),
                                     (  3,),
                                     (  3,))
        # A complex test
        self.check_message_to_parent((7,6,5,4,3),
                                     (  6,1,4,3),
                                     (1,1,5,4,1),
                                     (  6,5,1,3))
        # Test errors for inconsistent shapes
        self.assertRaises(ValueError, 
                          self.check_message_to_parent,
                          (3,),
                          (1,3,),
                          (3,),
                          (3,))
        ## self.assertRaises(ValueError, 
        ##                   self.check_message_to_parent,
        ##                   (3,),
        ##                   (3,),
        ##                   (1,3,),
        ##                   (3,))
        self.assertRaises(ValueError, 
                          self.check_message_to_parent,
                          (3,),
                          (1,3,),
                          (1,3,),
                          (3,))
        self.assertRaises(ValueError, 
                          self.check_message_to_parent,
                          (3,),
                          (4,),
                          (3,),
                          (3,))
        self.assertRaises(ValueError, 
                          self.check_message_to_parent,
                          (3,),
                          (3,),
                          (4,),
                          (3,))
        self.assertRaises(ValueError, 
                          self.check_message_to_parent,
                          (3,),
                          (4,),
                          (4,),
                          (3,))
        self.assertRaises(ValueError, 
                          self.check_message_to_parent,
                          (3,),
                          (4,),
                          (3,),
                          (1,))
        self.assertRaises(ValueError, 
                          self.check_message_to_parent,
                          (3,),
                          (3,),
                          (4,),
                          (1,))
        self.assertRaises(ValueError, 
                          self.check_message_to_parent,
                          (3,),
                          (4,),
                          (4,),
                          (1,))
        self.assertRaises(ValueError, 
                          self.check_message_to_parent,
                          (1,),
                          (4,),
                          (3,),
                          (1,))
        self.assertRaises(ValueError, 
                          self.check_message_to_parent,
                          (1,),
                          (3,),
                          (4,),
                          (1,))
        self.assertRaises(ValueError, 
                          self.check_message_to_parent,
                          (1,),
                          (4,),
                          (4,),
                          (1,))


    def test_compute_message(self):
        """
        Test the general sum-multiply function for message computations
        """

        self.assertAllClose(Node._compute_message(3,
                                                  plates_from=(),
                                                  plates_to=(),
                                                  ndim=0),
                            3)

        # Sum over one array
        self.assertAllClose(Node._compute_message([1, 2, 3],
                                                  plates_from=(3,),
                                                  plates_to=(),
                                                  ndim=0),
                            6)

        # Sum plates
        self.assertAllClose(Node._compute_message([1, 2, 3],
                                                  [4, 4, 4],
                                                  [5, 5, 5],
                                                  plates_from=(3,),
                                                  plates_to=(),
                                                  ndim=0),
                            20+40+60)

        # Do not sum plates
        self.assertAllClose(Node._compute_message([1, 2, 3],
                                                  [4, 4, 4],
                                                  [5, 5, 5],
                                                  plates_from=(3,),
                                                  plates_to=(3,),
                                                  ndim=0),
                            [20, 40, 60])

        # Give ndim
        self.assertAllClose(Node._compute_message([1, 2, 3],
                                                  [4, 4, 4],
                                                  [5, 5, 5],
                                                  plates_from=(),
                                                  plates_to=(),
                                                  ndim=1),
                            [20, 40, 60])

        # Broadcast plates_from
        self.assertAllClose(Node._compute_message(3,
                                                  4,
                                                  5,
                                                  plates_from=(3,),
                                                  plates_to=(),
                                                  ndim=0),
                            3 * (3*4*5))

        # Broadcast plates_to
        self.assertAllClose(Node._compute_message(3,
                                                  4,
                                                  5,
                                                  plates_from=(3,),
                                                  plates_to=(3,),
                                                  ndim=0),
                            3*4*5)

        # Different ndims
        self.assertAllClose(Node._compute_message([1, 2, 3],
                                                  [4],
                                                  5,
                                                  plates_from=(3,),
                                                  plates_to=(3,),
                                                  ndim=0),
                            [1*4*5, 2*4*5, 3*4*5])

        # Broadcasting dims for some arrays
        self.assertAllClose(Node._compute_message([1, 2, 3],
                                                  [4],
                                                  5,
                                                  plates_from=(),
                                                  plates_to=(),
                                                  ndim=1),
                            [1*4*5, 2*4*5, 3*4*5])

        # Bugfix: Check that plate keys are mapped correctly
        self.assertAllClose(
            Node._compute_message(
                [[1], [2], [3]],
                plates_from=(3,2),
                plates_to=(1,2),
                ndim=0
            ),
            [[6]]
        )

        # Bugfix: Check plate key mapping when plates_to is shorter than shape
        # of the array
        self.assertAllClose(
            Node._compute_message(
                [[1, 2, 3], [4, 5, 6]],
                plates_from=(2,3),
                plates_to=(3,),
                ndim=0
            ),
            [5, 7, 9]
        )

        # Complex example
        x1 = np.random.randn(5,4,1,2,1)
        x2 = np.random.randn(    1,2,1)
        x3 = np.random.randn(5,1,1,1,1)
        self.assertAllClose(Node._compute_message(x1, x2, x3,
                                                  plates_from=(6,5,4,3),
                                                  plates_to=(5,1,1),
                                                  ndim=2),
                            3*6*np.sum(x1*x2*x3, axis=(-4,-3), keepdims=True))

        pass


class TestSlice(misc.TestCase):

    def test_init(self):
        """
        Test the constructor of the X[..] node operator.
        """

        class MyNode(Node):
            _moments = Moments()
            _parent_moments = ()
            def _get_id_list(self):
                return []

        # Integer index
        X = MyNode(plates=(3,4), dims=((),))
        Y = X[2]
        self.assertEqual(Y.plates, (4,))
        X = MyNode(plates=(3,4), dims=((),))
        Y = X[(2,)]
        self.assertEqual(Y.plates, (4,))

        X = MyNode(plates=(3,4), dims=((),))
        Y = X[2,-4]
        self.assertEqual(Y.plates, ())

        X = MyNode(plates=(3,4,5), dims=((),))
        Y = X[2,1]
        self.assertEqual(Y.plates, (5,))

        # Full slices
        X = MyNode(plates=(3,4,5), dims=((),))
        Y = X[:,1,:]
        self.assertEqual(Y.plates, (3,5,))

        X = MyNode(plates=(3,4,5), dims=((),))
        Y = X[1,:,:]
        self.assertEqual(Y.plates, (4,5,))

        X = MyNode(plates=(3,4,5), dims=((),))
        Y = X[:,:,1]
        self.assertEqual(Y.plates, (3,4,))

        # Slice with step
        X = MyNode(plates=(9,), dims=((),))
        Y = X[::3]
        self.assertEqual(Y.plates, (3,))

        X = MyNode(plates=(10,), dims=((),))
        Y = X[::3]
        self.assertEqual(Y.plates, (4,))

        X = MyNode(plates=(11,), dims=((),))
        Y = X[::3]
        self.assertEqual(Y.plates, (4,))

        # Slice with a start value
        X = MyNode(plates=(10,), dims=((),))
        Y = X[3:]
        self.assertEqual(Y.plates, (7,))

        # Slice with an end value
        X = MyNode(plates=(10,), dims=((),))
        Y = X[:7]
        self.assertEqual(Y.plates, (7,))

        # Slice with only one element
        X = MyNode(plates=(10,), dims=((),))
        Y = X[6:7]
        self.assertEqual(Y.plates, (1,))

        # Slice starts out of range
        X = MyNode(plates=(10,), dims=((),))
        Y = X[-20:]
        self.assertEqual(Y.plates, (10,))

        # Slice ends out of range
        X = MyNode(plates=(10,), dims=((),))
        Y = X[:20]
        self.assertEqual(Y.plates, (10,))

        # Counter-intuitive: This slice is not empty
        X = MyNode(plates=(3,), dims=((),))
        Y = X[-4::4]
        self.assertEqual(Y.plates, (1,))
        
        # One ellipsis
        X = MyNode(plates=(3,4,5,6), dims=((),))
        Y = X[...,2,1]
        self.assertEqual(Y.plates, (3,4))

        X = MyNode(plates=(3,4,5,6), dims=((),))
        Y = X[2,...,1]
        self.assertEqual(Y.plates, (4,5))

        X = MyNode(plates=(3,4,5,6), dims=((),))
        Y = X[2,1,...]
        self.assertEqual(Y.plates, (5,6))

        # Multiple ellipsis
        X = MyNode(plates=(3,4,5), dims=((),))
        Y = X[...,2,...]
        self.assertEqual(Y.plates, (3,5))

        X = MyNode(plates=(3,4,5), dims=((),))
        Y = X[...,2,...,...]
        self.assertEqual(Y.plates, (4,5))

        X = MyNode(plates=(3,4,5), dims=((),))
        Y = X[...,...,...,...]
        self.assertEqual(Y.plates, (3,4,5))

        # New axis
        X = MyNode(plates=(3,), dims=((),))
        Y = X[None]
        self.assertEqual(Y.plates, (1,3))
        
        X = MyNode(plates=(3,), dims=((),))
        Y = X[:,None]
        self.assertEqual(Y.plates, (3,1))
        
        X = MyNode(plates=(3,4), dims=((),))
        Y = X[None,:,None,:]
        self.assertEqual(Y.plates, (1,3,1,4))

        #
        # Test errors
        #

        class Z:
            def __getitem__(self, obj):
                return obj

        # Invalid argument
        self.assertRaises(TypeError,
                          MyNode(plates=(3,),
                               dims=((),)).__getitem__,
                          Z()['a'])
        self.assertRaises(TypeError,
                          MyNode(plates=(3,),
                               dims=((),)).__getitem__,
                          Z()[[2,1]])

        # Too many indices
        self.assertRaises(IndexError,
                          MyNode(plates=(3,),
                               dims=((),)).__getitem__,
                          Z()[:,:])
        self.assertRaises(IndexError,
                          MyNode(plates=(3,),
                               dims=((),)).__getitem__,
                          Z()[...,...,...])

        # Index out of range
        self.assertRaises(IndexError,
                          MyNode(plates=(3,),
                               dims=((),)).__getitem__,
                          Z()[3])
        self.assertRaises(IndexError,
                          MyNode(plates=(3,),
                               dims=((),)).__getitem__,
                          Z()[-4])

        # Empty slice
        self.assertRaises(IndexError,
                          MyNode(plates=(3,),
                               dims=((),)).__getitem__,
                          Z()[3:])
        self.assertRaises(IndexError,
                          MyNode(plates=(3,),
                               dims=((),)).__getitem__,
                          Z()[:-3])
        
        pass

    def test_message_to_child(self):
        """
        Test message to child of X[..] node operator.
        """

        class DummyNode(Node):
            _moments = Moments()
            _parent_moments = (Moments(),)
            def __init__(self, u, **kwargs):
                self.u = u
                super().__init__(**kwargs)
            def _message_to_child(self):
                return self.u
            def _get_id_list(self):
                return []

        # Message not a reference to X.u but a copy of it
        X = DummyNode([np.random.randn(3)],
                      plates=(3,), 
                      dims=((),))
        Y = X[2]
        self.assertTrue(Y._message_to_child() is not X.u,
                        msg="Slice node operator sends a reference to the "
                            "node's moment list as a message instead of a copy "
                            "of the list.")
            
        # Integer indices
        X = DummyNode([np.random.randn(3,4)],
                      plates=(3,4), 
                      dims=((),))
        Y = X[2,1]
        self.assertMessageToChild(Y, [X.u[0][2,1]])

        # Too few integer indices
        X = DummyNode([np.random.randn(3,4)],
                      plates=(3,4), 
                      dims=((),))
        Y = X[2]
        self.assertMessageToChild(Y, [X.u[0][2]])

        # Integer for broadcasted moment
        X = DummyNode([np.random.randn(4)],
                      plates=(3,4), 
                      dims=((),))
        Y = X[2,1]
        self.assertMessageToChild(Y, [X.u[0][1]])
        X = DummyNode([np.random.randn(4,1)],
                      plates=(3,4), 
                      dims=((),))
        Y = X[2,1]
        self.assertMessageToChild(Y, [X.u[0][2,0]])


        # Ignore leading new axes
        X = DummyNode([np.random.randn(3)],
                      plates=(3,), 
                      dims=((),))
        Y = X[None,None,2]
        self.assertMessageToChild(Y, [X.u[0][2]])

        # Ignore new axes before missing+broadcasted plate axes
        X = DummyNode([np.random.randn(3)],
                      plates=(4,3,), 
                      dims=((),))
        Y = X[1,None,None,2]
        self.assertMessageToChild(Y, [X.u[0][2]])

        # New axes
        X = DummyNode([np.random.randn(3,4)],
                      plates=(3,4), 
                      dims=((),))
        Y = X[2,None,None,1]
        self.assertMessageToChild(Y, [X.u[0][2,None,None,1]])

        # New axes for broadcasted axes
        X = DummyNode([np.random.randn(4)],
                      plates=(3,4), 
                      dims=((),))
        Y = X[2,1,None,None]
        self.assertMessageToChild(Y, [X.u[0][1,None,None]])

        # Full slice
        X = DummyNode([np.random.randn(3,4)],
                      plates=(3,4), 
                      dims=((),))
        Y = X[:,2]
        self.assertMessageToChild(Y, [X.u[0][:,2]])

        # Slice with start
        X = DummyNode([np.random.randn(3,4)],
                      plates=(3,4), 
                      dims=((),))
        Y = X[1:,2]
        self.assertMessageToChild(Y, [X.u[0][1:,2]])

        # Slice with end
        X = DummyNode([np.random.randn(3,4)],
                      plates=(3,4), 
                      dims=((),))
        Y = X[:2,2]
        self.assertMessageToChild(Y, [X.u[0][:2,2]])

        # Slice with step
        X = DummyNode([np.random.randn(3,4)],
                      plates=(3,4), 
                      dims=((),))
        Y = X[::2,2]
        self.assertMessageToChild(Y, [X.u[0][::2,2]])

        # Slice for broadcasted axes
        X = DummyNode([np.random.randn(4)],
                      plates=(3,4), 
                      dims=((),))
        Y = X[0:2:2,2]
        self.assertMessageToChild(Y, [X.u[0][2]])
        X = DummyNode([np.random.randn(1,4)],
                      plates=(3,4), 
                      dims=((),))
        Y = X[0:2:2,2]
        self.assertMessageToChild(Y, [X.u[0][0:1,2]])

        # One ellipsis
        X = DummyNode([np.random.randn(3,4)],
                      plates=(3,4), 
                      dims=((),))
        Y = X[...,2]
        self.assertMessageToChild(Y, [X.u[0][...,2]])

        # Ellipsis over broadcasted axes
        X = DummyNode([np.random.randn(5,6)],
                      plates=(3,4,5,6), 
                      dims=((),))
        Y = X[1,...,2]
        self.assertMessageToChild(Y, [X.u[0][:,2]])
        X = DummyNode([np.random.randn(3,1,5,6)],
                      plates=(3,4,5,6), 
                      dims=((),))
        Y = X[1,...,2]
        self.assertMessageToChild(Y, [X.u[0][1,:,:,2]])

        # Multiple ellipsis
        X = DummyNode([np.random.randn(2,3,4,5)],
                      plates=(2,3,4,5), 
                      dims=((),))
        Y = X[...,2,...]
        self.assertMessageToChild(Y, [X.u[0][:,:,2,:]])

        # Ellipsis when dimensions
        X = DummyNode([np.random.randn(2,3,4)],
                      plates=(2,3), 
                      dims=((4,),))
        Y = X[...,2]
        self.assertMessageToChild(Y, [X.u[0][:,2,:]])

        # Indexing for multiple moments
        X = DummyNode([np.random.randn(2,3,4),
                       np.random.randn(2,3)],
                      plates=(2,3), 
                      dims=((4,),()))
        Y = X[1,1]
        self.assertMessageToChild(Y, [X.u[0][1,1],
                                      X.u[1][1,1]])

        pass

    def test_message_to_parent(self):
        """
        Test message to parent of X[..] node operator.
        """

        class ParentNode(Node):
            _moments = Moments()
            _parent_moments = ()
            def _get_id_list(self):
                return []
            
        class ChildNode(Node):
            _moments = Moments()
            _parent_moments = (Moments(),)
            def __init__(self, X, m, mask, **kwargs):
                super().__init__(X, **kwargs)
                self.m = m
                self.mask2 = mask
            def _message_to_parent(self, index, u_parent=None):
                return self.m
            def _mask_to_parent(self, index):
                return self.mask2
            def _get_id_list(self):
                return []

        # General broadcasting
        V = ParentNode(plates=(3,3,3),
                       dims=((),))
        X = V[...]
        m = [ np.random.randn(3,1) ]
        msg = [ np.zeros((1,3,1)) ]
        msg[0][:,:,:] = m[0]
        Y = ChildNode(X, m, True, dims=((),))
        X._update_mask()
        self.assertMessage(X._message_to_parent(0),
                           msg)

        # Integer indices
        V = ParentNode(plates=(3,4),
                 dims=((),))
        X = V[2,1]
        m = [np.random.randn()]
        msg = [ np.zeros((3,4)) ]
        msg[0][2,1] = m[0]
        Y = ChildNode(X, m, True, dims=((),))
        X._update_mask()
        self.assertMessage(X._message_to_parent(0),
                           msg)

        # Integer indices with broadcasting
        V = ParentNode(plates=(3,3),
                 dims=((),))
        X = V[2,2]
        m = [ np.random.randn(1) ]
        msg = [ np.zeros((3,3)) ]
        msg[0][2,2] = m[0]
        Y = ChildNode(X, m, True, dims=((),))
        X._update_mask()
        self.assertMessage(X._message_to_parent(0),
                           msg)

        # Slice indices
        V = ParentNode(plates=(2,3,4,5),
                 dims=((),))
        X = V[:,:2,1:,::2]
        m = [np.random.randn(2,2,3,3)]
        msg = [ np.zeros((2,3,4,5)) ]
        msg[0][:,:2,1:,::2] = m[0]
        Y = ChildNode(X, m, True, dims=((),))
        X._update_mask()
        self.assertMessage(X._message_to_parent(0),
                           msg)

        # Full slice with broadcasting
        V = ParentNode(plates=(2,3),
                 dims=((),))
        X = V[:,:]
        m = [np.random.randn(1)]
        msg = [ np.zeros((1,1)) ]
        msg[0][:] = m[0]
        Y = ChildNode(X, m, True, dims=((),))
        X._update_mask()
        self.assertMessage(X._message_to_parent(0),
                           msg)

        # Start slice with broadcasting
        V = ParentNode(plates=(3,3,3,3),
                 dims=((),))
        X = V[0:,1:,-2:,-3:]
        m = [np.random.randn(1,1)]
        msg = [ np.zeros((1,3,3,1)) ]
        msg[0][:,1:,-2:,:] = m[0]
        Y = ChildNode(X, m, True, dims=((),))
        X._update_mask()
        self.assertMessage(X._message_to_parent(0),
                           msg)

        # End slice with broadcasting
        V = ParentNode(plates=(3,3,3,3),
                 dims=((),))
        X = V[:2,:3,:4,:-1]
        m = [np.random.randn(1,1)]
        msg = [ np.zeros((3,1,1,3)) ]
        msg[0][:2,:,:,:-1] = m[0]
        Y = ChildNode(X, m, True, dims=((),))
        X._update_mask()
        self.assertMessage(X._message_to_parent(0),
                           msg)

        # Step slice with broadcasting
        V = ParentNode(plates=(3,3,1),
                 dims=((),))
        X = V[::1,::2,::2]
        m = [np.random.randn(1)]
        msg = [ np.zeros((1,3,1)) ]
        msg[0][:,::2,:] = m[0]
        Y = ChildNode(X, m, True, dims=((),))
        X._update_mask()
        self.assertMessage(X._message_to_parent(0),
                           msg)

        # Ellipsis
        V = ParentNode(plates=(3,3,3),
                 dims=((),))
        X = V[...,0]
        m = [np.random.randn(3,3)]
        msg = [ np.zeros((3,3,3)) ]
        msg[0][:,:,0] = m[0]
        Y = ChildNode(X, m, True, dims=((),))
        X._update_mask()
        self.assertMessage(X._message_to_parent(0),
                           msg)
        
        # New axes
        V = ParentNode(plates=(3,3),
                 dims=((),))
        X = V[None,:,None,None,:]
        m = [np.random.randn(1,3,1,1,3)]
        msg = [ np.zeros((3,3)) ]
        msg[0][:,:] = m[0][0,:,0,0,:]
        Y = ChildNode(X, m, True, dims=((),))
        X._update_mask()
        self.assertMessage(X._message_to_parent(0),
                           msg)

        # New axes with broadcasting
        V = ParentNode(plates=(3,3),
                 dims=((),))
        X = V[None,:,None,:,None]
        m = [np.random.randn(1,3,1)]
        msg = [ np.zeros((1,3)) ]
        msg[0][:,:] = m[0][0,:,0]
        Y = ChildNode(X, m, True, dims=((),))
        X._update_mask()
        self.assertMessage(X._message_to_parent(0),
                           msg)

        # Multiple messages
        V = ParentNode(plates=(3,),
                 dims=((),()))
        X = V[:]
        m = [np.random.randn(3),
             np.random.randn(3)]
        msg = [ np.zeros((3)), np.zeros((3)) ]
        msg[0][:] = m[0][:]
        msg[1][:] = m[1][:]
        Y = ChildNode(X, m, True, dims=((),()))
        X._update_mask()
        self.assertMessage(X._message_to_parent(0),
                           msg)

        # Non-scalar variables
        V = ParentNode(plates=(2,3),
                 dims=((4,),))
        X = V[...]
        m = [np.random.randn(2,3,4)]
        msg = [ np.zeros((2,3,4)) ]
        msg[0][:,:,:] = m[0][:,:,:]
        Y = ChildNode(X, m, True, dims=((4,),))
        X._update_mask()
        self.assertMessage(X._message_to_parent(0),
                           msg)

        # Missing values
        V = ParentNode(plates=(3,3,3),
                 dims=((3,),))
        X = V[:,0,::2,None]
        m = [np.random.randn(3,2,1,3)]
        # mask shape: (3, 2, 1)
        mask = np.array([ [[True],  [False]],
                          [[False], [False]],
                          [[False], [True]] ])
        msg = [ np.zeros((3,3,3,3)) ]
        msg[0][:,0,::2,:] = (m[0] * mask[...,None])[:,:,0,:]
        Y = ChildNode(X, m, mask, dims=((3,),))
        X._update_mask()
        self.assertMessage(X._message_to_parent(0),
                           msg)

        # Found bug: int index after slice
        V = ParentNode(plates=(3,3),
                 dims=((),))
        X = V[:,0]
        m = [np.random.randn(3)]
        msg = [ np.zeros((3,3)) ]
        msg[0][:,0] = m[0]
        Y = ChildNode(X, m, True, dims=((),))
        X._update_mask()
        self.assertMessage(X._message_to_parent(0),
                           msg)

        # Found bug: message requires reshaping after reverse indexing
        

        pass



################################################################################
# Copyright (C) 2015 Jaakko Luttinen
#
# This file is licensed under the MIT License.
################################################################################

import numpy as np

from .deterministic import Deterministic
from .node import Moments
from bayespy.utils import misc


class Take(Deterministic):
    """
    Choose elements/sub-arrays along a plate axis

    Basically, applies `np.take` on a plate axis. Allows advanced mapping of
    plates.

    Parameters
    ----------
    node : Node
        A node to apply the take operation on.
    indices : array of integers
        Plate elements to pick along a plate axis.
    plate_axis : int (negative)
        The plate axis to pick elements from (default: -1).

    See also
    --------
    numpy.take

    Examples
    --------

    >>> from bayespy.nodes import Gamma, Take
    >>> alpha = Gamma([1, 2, 3], [1, 1, 1])
    >>> x = Take(alpha, [1, 1, 2, 2, 1, 0])
    >>> x.get_moments()[0]
    array([2., 2., 3., 3., 2., 1.])
    """


    def __init__(self, node, indices, plate_axis=-1, **kwargs):
        self._moments = node._moments
        self._parent_moments = (node._moments,)
        self._indices = np.array(indices)
        self._plate_axis = plate_axis
        self._original_length = node.plates[plate_axis]

        # Validate arguments
        if not misc.is_scalar_integer(plate_axis):
            raise ValueError("Plate axis must be integer")
        if plate_axis >= 0:
            raise ValueError("plate_axis must be negative index")
        if plate_axis < -len(node.plates):
            raise ValueError("plate_axis out of bounds")
        if not issubclass(self._indices.dtype.type, np.integer):
            raise ValueError("Indices must be integers")
        if (np.any(self._indices < -self._original_length) or
            np.any(self._indices >= self._original_length)):
            raise ValueError("Index out of bounds")

        super().__init__(node, dims=node.dims, **kwargs)


    def _compute_moments(self, u_parent):
        u = []
        for (ui, dimi) in zip(u_parent, self.dims):
            axis = self._plate_axis - len(dimi)
            # Just in case the taken axis is using broadcasting and has unit
            # length in u_parent, force it to have the correct length along the
            # axis in order to avoid errors in np.take.
            broadcaster = np.ones((self._original_length,) + (-axis-1)*(1,))
            u.append(np.take(ui*broadcaster, self._indices, axis=axis))
        return u


    def _compute_message_to_parent(self, index, m_child, u_parent):

        m = [
            misc.put_simple(
                mi,
                self._indices,
                axis=self._plate_axis-len(dimi),
                length=self._original_length,
            )
            for (mi, dimi) in zip(m_child, self.dims)
        ]
        return m


    def _compute_weights_to_parent(self, index, weights):

        return misc.put_simple(
            weights,
            self._indices,
            axis=self._plate_axis,
            length=self._original_length,
        )


    def _compute_plates_to_parent(self, index, plates):

        # Number of axes created by take operation
        N = np.ndim(self._indices)

        if self._plate_axis >= 0:
            raise RuntimeError("Plate axis should be negative")

        end_before = self._plate_axis - N + 1
        start_after = self._plate_axis + 1

        if end_before == 0:
            return plates + (self._original_length,)
        elif start_after == 0:
            return plates[:end_before] + (self._original_length,)

        return (plates[:end_before]
                + (self._original_length,)
                + plates[start_after:])


    def _compute_plates_from_parent(self, index, parent_plates):

        plates = parent_plates[:self._plate_axis] + np.shape(self._indices)
        if self._plate_axis != -1:
            plates = plates + parent_plates[(self._plate_axis+1):]
        return plates


    def _compute_plates_multiplier_from_parent(self, index, parent_multiplier):
        if any(p != 1 for p in parent_multiplier):
            raise NotImplementedError(
                "Take node doesn't yet support plate multipliers {0}"
                .format(parent_multiplier)
            )

        return parent_multiplier



################################################################################
# Copyright (C) 2015 Jaakko Luttinen
#
# This file is licensed under the MIT License.
################################################################################


"""
Unit tests for `vmp` module.
"""

import numpy as np

from scipy import special

from numpy import testing

from bayespy.nodes import (Gaussian, 
                           GaussianARD,
                           GaussianGamma,
                           Gamma,
                           Wishart)

from ..vmp import VB

from bayespy.utils import misc
from bayespy.utils import linalg
from bayespy.utils import random

from bayespy.utils.misc import TestCase

class TestVB(TestCase):

    def test_annealing(self):

        X = GaussianARD(3, 4)
        X.initialize_from_parameters(-1, 6)

        Q = VB(X)
        Q.set_annealing(0.1)

        #
        # Check that the gradient is correct
        #

        # Initial parameters 
        phi0 = X.phi
        # Gradient
        rg = X.get_riemannian_gradient()
        g = X.get_gradient(rg)
        # Numerical gradient of the first parameter
        eps = 1e-6
        p0 = X.get_parameters()
        l0 = Q.compute_lowerbound(ignore_masked=False)
        g_num = [(), ()]
        e = eps
        p1 = p0[0] + e
        X.set_parameters([p1, p0[1]])
        l1 = Q.compute_lowerbound(ignore_masked=False)
        g_num[0] = (l1 - l0) / eps
        # Numerical gradient of the second parameter
        p1 = p0[1] + e
        X.set_parameters([p0[0], p1])
        l1 = Q.compute_lowerbound(ignore_masked=False)
        g_num[1] = (l1 - l0) / (eps)
        # Check
        self.assertAllClose(g[0],
                            g_num[0])
        self.assertAllClose(g[1],
                            g_num[1])

        #
        # Gradient should be zero after updating
        #

        X.update()
        # Initial parameters 
        phi0 = X.phi
        # Numerical gradient of the first parameter
        eps = 1e-8
        p0 = X.get_parameters()
        l0 = Q.compute_lowerbound(ignore_masked=False)
        g_num = [(), ()]
        e = eps
        p1 = p0[0] + e
        X.set_parameters([p1, p0[1]])
        l1 = Q.compute_lowerbound(ignore_masked=False)
        g_num[0] = (l1 - l0) / eps
        # Numerical gradient of the second parameter
        p1 = p0[1] + e
        X.set_parameters([p0[0], p1])
        l1 = Q.compute_lowerbound(ignore_masked=False)
        g_num[1] = (l1 - l0) / (eps)
        # Check
        self.assertAllClose(0,
                            g_num[0],
                            atol=1e-5)
        self.assertAllClose(0,
                            g_num[1],
                            atol=1e-5)

        # Not at the optimum
        X.initialize_from_parameters(-1, 6)
        # Initial parameters 
        phi0 = X.phi
        # Gradient
        g = X.get_riemannian_gradient()
        # Parameters after VB-EM update
        X.update()
        phi1 = X.phi
        # Check
        self.assertAllClose(g[0],
                            phi1[0] - phi0[0])
        self.assertAllClose(g[1],
                            phi1[1] - phi0[1])
        

        pass



################################################################################
# Copyright (C) 2014 Jaakko Luttinen
#
# This file is licensed under the MIT License.
################################################################################


"""
Unit tests for `categorical` module.
"""

import warnings

import numpy as np
import scipy

from bayespy.nodes import (Categorical,
                           Dirichlet,
                           Mixture,
                           Gamma)

from bayespy.utils import random
from bayespy.utils import misc

from bayespy.utils.misc import TestCase


class TestCategorical(TestCase):
    """
    Unit tests for Categorical node
    """

    
    def test_init(self):
        """
        Test the creation of categorical nodes.
        """

        # Some simple initializations
        X = Categorical([0.1, 0.3, 0.6])
        X = Categorical(Dirichlet([5,4,3]))

        # Check that plates are correct
        X = Categorical([0.1, 0.3, 0.6], plates=(3,4))
        self.assertEqual(X.plates,
                         (3,4))
        X = Categorical(0.25*np.ones((2,3,4)))
        self.assertEqual(X.plates,
                         (2,3))
        X = Categorical(Dirichlet([2,1,9], plates=(3,4)))
        self.assertEqual(X.plates,
                         (3,4))
        

        # Probabilities not a vector
        self.assertRaises(ValueError,
                          Categorical,
                          0.5)

        # Invalid probability
        self.assertRaises(ValueError,
                          Categorical,
                          [-0.5, 1.5],
                          n=10)
        self.assertRaises(ValueError,
                          Categorical,
                          [0.5, 1.5],
                          n=10)

        # Inconsistent plates
        self.assertRaises(ValueError,
                          Categorical,
                          0.25*np.ones((2,4)),
                          plates=(3,),
                          n=10)

        # Explicit plates too small
        self.assertRaises(ValueError,
                          Categorical,
                          0.25*np.ones((2,4)),
                          plates=(1,),
                          n=10)

        pass

    
    def test_moments(self):
        """
        Test the moments of categorical nodes.
        """

        # Simple test
        X = Categorical([0.7,0.2,0.1])
        u = X._message_to_child()
        self.assertEqual(len(u), 1)
        self.assertAllClose(u[0],
                            [0.7,0.2,0.1])

        # Test plates in p
        p = np.random.dirichlet([1,1], size=3)
        X = Categorical(p)
        u = X._message_to_child()
        self.assertAllClose(u[0],
                            p)
        
        # Test with Dirichlet prior
        P = Dirichlet([7, 3])
        logp = P._message_to_child()[0]
        p0 = np.exp(logp[0]) / (np.exp(logp[0]) + np.exp(logp[1]))
        p1 = np.exp(logp[1]) / (np.exp(logp[0]) + np.exp(logp[1]))
        X = Categorical(P)
        u = X._message_to_child()
        p = np.array([p0, p1])
        self.assertAllClose(u[0],
                            p)

        # Test with broadcasted plates
        P = Dirichlet([7, 3], plates=(10,))
        X = Categorical(P)
        u = X._message_to_child()
        self.assertAllClose(u[0] * np.ones(X.get_shape(0)),
                            p*np.ones((10,1)))

        pass


    def test_observed(self):
        """
        Test observed categorical nodes
        """

        # Single observation
        X = Categorical([0.7,0.2,0.1])
        X.observe(2)
        u = X._message_to_child()
        self.assertAllClose(u[0],
                            [0,0,1])

        # One plate axis
        X = Categorical([0.7,0.2,0.1], plates=(2,))
        X.observe([2,1])
        u = X._message_to_child()
        self.assertAllClose(u[0],
                            [[0,0,1],
                             [0,1,0]])

        # Several plate axes
        X = Categorical([0.7,0.1,0.1,0.1], plates=(2,3,))
        X.observe([[2,1,1],
                   [0,2,3]])
        u = X._message_to_child()
        self.assertAllClose(u[0],
                            [ [[0,0,1,0],
                               [0,1,0,0],
                               [0,1,0,0]],
                              [[1,0,0,0],
                               [0,0,1,0],
                               [0,0,0,1]] ])

        # Check invalid observations
        X = Categorical([0.7,0.2,0.1])
        self.assertRaises(ValueError,
                          X.observe,
                          -1)
        self.assertRaises(ValueError,
                          X.observe,
                          3)
        self.assertRaises(ValueError,
                          X.observe,
                          1.5)

        pass

    
    def test_constant(self):
        """
        Test constant categorical nodes
        """

        # Basic test
        Y = Mixture(2, Gamma, [1, 2, 3], [1, 1, 1])
        u = Y._message_to_child()
        self.assertAllClose(u[0],
                            3/1)

        # Test with one plate axis
        alpha = [[1, 2, 3],
                 [4, 5, 6]]
        Y = Mixture([2, 1], Gamma, alpha, 1)
        u = Y._message_to_child()
        self.assertAllClose(u[0],
                            [3, 5])

        # Test with two plate axes
        alpha = [ [[1, 2, 3],
                   [4, 5, 6]],
                  [[7, 8, 9],
                   [10, 11, 12]] ]
        Y = Mixture([[2, 1], [0, 2]], Gamma, alpha, 1)
        u = Y._message_to_child()
        self.assertAllClose(u[0],
                            [[3, 5],
                             [7, 12]])

        pass


    def test_initialization(self):
        """
        Test initialization of categorical nodes
        """

        # Test initialization from random
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", RuntimeWarning)

            Z = Categorical([[0.0, 1.0, 0.0],
                             [0.0, 0.0, 1.0]])
            Z.initialize_from_random()
            u = Z._message_to_child()
            self.assertAllClose(u[0],
                                [[0, 1, 0],
                                 [0, 0, 1]])

        pass


    def test_gradient(self):
        """
        Check the Euclidean gradient of the categorical node
        """

        Z = Categorical([[0.3, 0.5, 0.2], [0.1, 0.6, 0.3]])
        Y = Mixture(Z, Gamma, [2, 3, 4], [5, 6, 7])
        Y.observe([4.2, 0.2])
        def f(x):
            Z.set_parameters([np.reshape(x, Z.get_shape(0))])
            return Z.lower_bound_contribution() + Y.lower_bound_contribution()
        def df(x):
            Z.set_parameters([np.reshape(x, Z.get_shape(0))])
            g = Z.get_riemannian_gradient()
            return Z.get_gradient(g)[0]
        x0 = np.ravel(np.log([[2, 3, 7], [0.1, 3, 1]]))
        self.assertAllClose(
            misc.gradient(f, x0),
            np.ravel(df(x0))
        )

        pass



################################################################################
# Copyright (C) 2013-2014 Jaakko Luttinen
#
# This file is licensed under the MIT License.
################################################################################


import numpy as np
import functools

from bayespy.utils import misc

"""
This module contains a sketch of a new implementation of the framework.
"""

def message_sum_multiply(plates_parent, dims_parent, *arrays):
    """
    Compute message to parent and sum over plates.

    Divide by the plate multiplier.
    """
    # The shape of the full message
    shapes = [np.shape(array) for array in arrays]
    shape_full = misc.broadcasted_shape(*shapes)
    # Find axes that should be summed
    shape_parent = plates_parent + dims_parent
    sum_axes = misc.axes_to_collapse(shape_full, shape_parent)
    # Compute the multiplier for cancelling the
    # plate-multiplier.  Because we are summing over the
    # dimensions already in this function (for efficiency), we
    # need to cancel the effect of the plate-multiplier
    # applied in the message_to_parent function.
    r = 1
    for j in sum_axes:
        if j >= 0 and j < len(plates_parent):
            r *= shape_full[j]
        elif j < 0 and j < -len(dims_parent):
            r *= shape_full[j]
    # Compute the sum-product
    m = misc.sum_multiply(*arrays,
                          axis=sum_axes,
                          sumaxis=True,
                          keepdims=True) / r
    # Remove extra axes
    m = misc.squeeze_to_dim(m, len(shape_parent))
    return m


class Moments():
    """
    Base class for defining the expectation of the sufficient statistics.

    The benefits:

      * Write statistic-specific features in one place only. For instance,
        covariance from Gaussian message.

      * Different nodes may have identically defined statistic so you need to
        implement related features only once. For instance, Gaussian and
        GaussianARD differ on the prior but the moments are the same.

      * General processing nodes which do not change the type of the moments may
        "inherit" the features from the parent node. For instance, slicing
        operator.

      * Conversions can be done easily in both of the above cases if the message
        conversion is defined in the moments class. For instance,
        GaussianMarkovChain to Gaussian and VaryingGaussianMarkovChain to
        Gaussian.
    """

    _converters = {}


    class NoConverterError(Exception):
        pass


    def get_instance_converter(self, **kwargs):
        """Default converter within a moments class is an identity.

        Override this method when moment class instances are not identical if
        they have different attributes.

        """
        if len(kwargs) > 0:
            raise NotImplementedError(
                "get_instance_converter not implemented for class {0}"
                .format(self.__class__.__name__)
            )
        return None


    def get_instance_conversion_kwargs(self):
        """
        Override this method when moment class instances are not identical if
        they have different attributes.
        """
        return {}


    @classmethod
    def add_converter(cls, moments_to, converter):
        cls._converters = cls._converters.copy()
        cls._converters[moments_to] = converter
        return


    def get_converter(self, moments_to):
        """
        Finds conversion to another moments type if possible.

        Note that a conversion from moments A to moments B may require
        intermediate conversions.  For instance: A->C->D->B.  This method finds
        the path which uses the least amount of conversions and returns that
        path as a single conversion.  If no conversion path is available, an
        error is raised.

        The search algorithm starts from the original moments class and applies
        all possible converters to get a new list of moments classes. This list
        is extended by adding recursively all parent classes because their
        converters are applicable. Then, all possible converters are applied to
        this list to get a new list of current moments classes. This is iterated
        until the algorithm hits the target moments class or its subclass.
        """

        # Check if there is no need for a conversion
        #
        # TODO/FIXME: This isn't sufficient. Moments can have attributes that
        # make them incompatible (e.g., ndim in GaussianMoments).
        if isinstance(self, moments_to):
            return lambda X: X

        # Initialize variables
        visited = set()
        visited.add(self.__class__)
        converted_list = [(self.__class__, [])]

        # Each iteration step consists of two parts:
        # 1) form a set of the current classes and all their parent classes
        #    recursively
        # 2) from the current set, apply possible conversions to get a new set
        #    of classes
        # Repeat these two steps until in step (1) you hit the target class.

        while len(converted_list) > 0:
            # Go through all parents recursively so we can then use all
            # converters that are available
            current_list = []
            for (moments_class, converter_path) in converted_list:
                if issubclass(moments_class, moments_to):
                    # Shortest conversion path found, return the resulting total
                    # conversion function
                    return misc.composite_function(converter_path)
                current_list.append((moments_class, converter_path))
                parents = list(moments_class.__bases__)
                for parent in parents:
                    # Recursively add parents
                    for p in parent.__bases__:
                        if isinstance(p, Moments):
                            parents.append(p)
                    # Add un-visited parents
                    if issubclass(parent, Moments) and parent not in visited:
                        visited.add(parent)
                        current_list.append((parent, converter_path))

            # Find all converters and extend the converter paths
            converted_list = []
            for (moments_class, converter_path) in current_list:
                for (conv_mom_cls, conv) in moments_class._converters.items():
                    if conv_mom_cls not in visited:
                        visited.add(conv_mom_cls)
                        converted_list.append((conv_mom_cls,
                                               converter_path + [conv]))

        raise self.NoConverterError("No conversion defined from %s to %s"
                                    % (self.__class__.__name__,
                                       moments_to.__name__))


    def compute_fixed_moments(self, x):
        # This method can't be static because the computation of the moments may
        # depend on, for instance, ndim in Gaussian arrays.
        raise NotImplementedError("compute_fixed_moments not implemented for "
                                  "%s"
                                  % (self.__class__.__name__))


    @classmethod
    def from_values(cls, x):
        raise NotImplementedError("from_values not implemented "
                                  "for %s"
                                  % (cls.__name__))

def ensureparents(func):
    @functools.wraps(func)
    def wrapper(self, *parents, **kwargs):
        # Convert parents to proper nodes
        if self._parent_moments is None:
            raise ValueError(
                "Parent moments must be defined for {0}"
                .format(self.__class__.__name__)
            )
        parents = [
            Node._ensure_moments(
                parent,
                moments.__class__,
                **moments.get_instance_conversion_kwargs()
            )
            for (parent, moments) in zip(parents, self._parent_moments)
        ]
        # parents = list(parents)
        # for (ind, parent) in enumerate(parents):
        #     parents[ind] = self._ensure_moments(parent,
        #                                         self._parent_moments[ind])
        # Run the function
        return func(self, *parents, **kwargs)

    return wrapper


class Node():
    """
    Base class for all nodes.

    mask
    dims
    plates
    parents
    children
    name

    Sub-classes must implement:
    1. For computing the message to children:
       get_moments(self):
    2. For computing the message to parents:
       _get_message_and_mask_to_parent(self, index)

    Sub-classes may need to re-implement:
    1. If they manipulate plates:
       _compute_weights_to_parent(index, weights)
       _plates_to_parent(self, index)
       _plates_from_parent(self, index)
    """

    # These are objects of the _parent_moments_class. If the default way of
    # creating them is not correct, write your own creation code.
    _moments = None
    _parent_moments = None
    plates = None

    _id_counter = 0

    @ensureparents
    def __init__(self, *parents, dims=None, plates=None, name="",
                 notify_parents=True, plotter=None, plates_multiplier=None,
                 allow_dependent_parents=False):

        self.parents = parents
        self.dims = dims
        self.name = name
        self._plotter = plotter

        if not allow_dependent_parents:
            parent_id_list = []
            for parent in parents:
                parent_id_list = parent_id_list + list(parent._get_id_list())
            if len(parent_id_list) != len(set(parent_id_list)):
                raise ValueError("Parent nodes are not independent")

        # Inform parent nodes
        if notify_parents:
            for (index,parent) in enumerate(self.parents):
                parent._add_child(self, index)

        # Check plates
        parent_plates = [self._plates_from_parent(index)
                         for index in range(len(self.parents))]
        if any(p is None for p in parent_plates):
            raise ValueError("Method _plates_from_parent returned None")

        # Get and validate the plates for this node
        plates = self._total_plates(plates, *parent_plates)
        if self.plates is None:
            self.plates = plates

        # By default, ignore all plates
        self.mask = np.array(False)

        # Children
        self.children = set()

        # Get and validate the plate multiplier
        parent_plates_multiplier = [self._plates_multiplier_from_parent(index)
                                   for index in range(len(self.parents))]
        #if plates_multiplier is None:
        #    plates_multiplier = parent_plates_multiplier
        plates_multiplier = self._total_plates(plates_multiplier,
                                              *parent_plates_multiplier)
        self.plates_multiplier = plates_multiplier


    def get_pdf_nodes(self):
        return tuple(
            node
            for (child, _) in self.children
            for node in child._get_pdf_nodes_conditioned_on_parents()
        )


    def _get_pdf_nodes_conditioned_on_parents(self):
        return self.get_pdf_nodes()


    def _get_id_list(self):
        """
        Returns the stochastic ID list.

        This method is used to check that same stochastic nodes are not direct
        parents of a node several times. It is only valid if there are
        intermediate stochastic nodes.

        To put it another way: each ID corresponds to one factor q(..) in the
        posterior approximation. Different IDs mean different factors, thus they
        mean independence. The parents must have independent factors.

        Stochastic nodes should return their unique ID. Deterministic nodes
        should return the IDs of their parents. Constant nodes should return
        empty list of IDs.
        """
        raise NotImplementedError()


    @classmethod
    def _total_plates(cls, plates, *parent_plates):
        if plates is None:
            # By default, use the minimum number of plates determined
            # from the parent nodes
            try:
                return misc.broadcasted_shape(*parent_plates)
            except ValueError:
                raise ValueError(
                    "The plates of the parents do not broadcast: {0}".format(
                        parent_plates
                    )
                )
        else:
            # Check that the parent_plates are a subset of plates.
            for (ind, p) in enumerate(parent_plates):
                if not misc.is_shape_subset(p, plates):
                    raise ValueError("The plates %s of the parents "
                                     "are not broadcastable to the given "
                                     "plates %s."
                                     % (p,
                                        plates))
            return plates


    @staticmethod
    def _ensure_moments(node, moments_class, **kwargs):
        try:
            converter = node._moments.get_converter(moments_class)
        except AttributeError:
            from .constant import Constant
            return Constant(
                moments_class.from_values(node, **kwargs),
                node
            )
        else:
            node = converter(node)
            converter = node._moments.get_instance_converter(**kwargs)
            if converter is not None:
                from .converters import NodeConverter
                return NodeConverter(converter, node)
            return node


    def _compute_plates_to_parent(self, index, plates):
        # Sub-classes may want to overwrite this if they manipulate plates
        return plates


    def _compute_plates_from_parent(self, index, plates):
        # Sub-classes may want to overwrite this if they manipulate plates
        return plates


    def _compute_plates_multiplier_from_parent(self, index, plates_multiplier):
        # TODO/FIXME: How to handle this properly?
        return plates_multiplier


    def _plates_to_parent(self, index):
        return self._compute_plates_to_parent(index, self.plates)


    def _plates_from_parent(self, index):
        return self._compute_plates_from_parent(index,
                                                self.parents[index].plates)


    def _plates_multiplier_from_parent(self, index):
        return self._compute_plates_multiplier_from_parent(
            index,
            self.parents[index].plates_multiplier
        )


    @property
    def plates_multiplier(self):
        """ Plate multiplier is applied to messages to parents """
        return self.__plates_multiplier


    @plates_multiplier.setter
    def plates_multiplier(self, value):
        # TODO/FIXME: Check that multiplier is consistent with plates
        self.__plates_multiplier = value
        return


    def get_shape(self, ind):
        return self.plates + self.dims[ind]

    def _add_child(self, child, index):
        """
        Add a child node.

        Parameters
        ----------
        child : node
        index : int
           The parent index of this node for the child node.
           The child node recognizes its parents by their index
           number.
        """
        self.children.add((child, index))

    def _remove_child(self, child, index):
        """
        Remove a child node.
        """
        self.children.remove((child, index))

    def get_mask(self):
        return self.mask

    ## def _get_message_mask(self):
    ##     return self.mask

    def _set_mask(self, mask):
        # Sub-classes may overwrite this method if they have some other masks to
        # be combined (for instance, observation mask)
        self.mask = mask

    def _update_mask(self):
        # Combine masks from children
        mask = np.array(False)
        for (child, index) in self.children:
            mask = np.logical_or(mask, child._mask_to_parent(index))
        # Set the mask of this node
        self._set_mask(mask)
        if not misc.is_shape_subset(np.shape(self.mask), self.plates):

            raise ValueError("The mask of the node %s has updated "
                             "incorrectly. The plates in the mask %s are not a "
                             "subset of the plates of the node %s."
                             % (self.name,
                                np.shape(self.mask),
                                self.plates))

        # Tell parents to update their masks
        for parent in self.parents:
            parent._update_mask()


    def _compute_weights_to_parent(self, index, weights):
        """Compute the mask used for messages sent to parent[index].

        The mask tells which plates in the messages are active. This method is
        used for obtaining the mask which is used to set plates in the messages
        to parent to zero.

        Sub-classes may want to overwrite this method if they do something to
        plates so that the mask is somehow altered.

        """
        return weights


    def _mask_to_parent(self, index):
        """
        Get the mask with respect to parent[index].

        The mask tells which plate connections are active. The mask is "summed"
        (logical or) and reshaped into the plate shape of the parent. Thus, it
        can't be used for masking messages, because some plates have been summed
        already. This method is used for propagating the mask to parents.
        """
        mask = self._compute_weights_to_parent(index, self.mask) != 0

        # Check the shape of the mask
        plates_to_parent = self._plates_to_parent(index)
        if not misc.is_shape_subset(np.shape(mask), plates_to_parent):
            raise ValueError("In node %s, the mask being sent to "
                             "parent[%d] (%s) has invalid shape: The shape of "
                             "the mask %s is not a sub-shape of the plates of "
                             "the node with respect to the parent %s. It could "
                             "be that this node (%s) is manipulating plates "
                             "but has not overwritten the method "
                             "_compute_weights_to_parent."
                             % (self.name,
                                index,
                                self.parents[index].name,
                                np.shape(mask),
                                plates_to_parent,
                                self.__class__.__name__))

        # "Sum" (i.e., logical or) over the plates that have unit length in
        # the parent node.
        parent_plates = self.parents[index].plates
        s = misc.axes_to_collapse(np.shape(mask), parent_plates)
        mask = np.any(mask, axis=s, keepdims=True)
        mask = misc.squeeze_to_dim(mask, len(parent_plates))
        return mask

    def _message_to_child(self):

        u = self.get_moments()

        # Debug: Check that the message has appropriate shape
        for (ui, dim) in zip(u, self.dims):
            ndim = len(dim)
            if ndim > 0:
                if np.shape(ui)[-ndim:] != dim:
                    raise RuntimeError(
                        "A bug found by _message_to_child for %s: "
                        "The variable axes of the moments %s are not equal to "
                        "the axes %s defined by the node %s. A possible reason "
                        "is that the plates of the node are inferred "
                        "incorrectly from the parents, and the method "
                        "_plates_from_parents should be implemented."
                        % (self.__class__.__name__,
                           np.shape(ui)[-ndim:],
                           dim,
                           self.name))
                if not misc.is_shape_subset(np.shape(ui)[:-ndim],
                                            self.plates):
                    raise RuntimeError(
                        "A bug found by _message_to_child for %s: "
                        "The plate axes of the moments %s are not a subset of "
                        "the plate axes %s defined by the node %s."
                        % (self.__class__.__name__,
                           np.shape(ui)[:-ndim],
                           self.plates,
                           self.name))
            else:
                if not misc.is_shape_subset(np.shape(ui), self.plates):
                    raise RuntimeError(
                        "A bug found by _message_to_child for %s: "
                        "The plate axes of the moments %s are not a subset of "
                        "the plate axes %s defined by the node %s."
                        % (self.__class__.__name__,
                           np.shape(ui),
                           self.plates,
                           self.name))
        return u

    def _message_to_parent(self, index, u_parent=None):

        # Compute the message, check plates, apply mask and sum over some plates
        if index >= len(self.parents):
            raise ValueError("Parent index larger than the number of parents")

        # Compute the message and mask
        (m, mask) = self._get_message_and_mask_to_parent(index, u_parent=u_parent)
        mask = misc.squeeze(mask)

        # Plates in the mask
        plates_mask = np.shape(mask)

        # The parent we're sending the message to
        parent = self.parents[index]

        # Plates with respect to the parent
        plates_self = self._plates_to_parent(index)

        # Plate multiplier of the parent
        multiplier_parent = self._plates_multiplier_from_parent(index)

        # Check if m is a logpdf function (for black-box variational inference)
        if callable(m):
            return m

            def m_function(*args):
                lpdf = m(*args)
                # Log pdf only contains plate axes!
                plates_m = np.shape(lpdf)
                r = (self.broadcasting_multiplier(plates_self,
                                                  plates_m,
                                                  plates_mask,
                                                  parent.plates) *
                     self.broadcasting_multiplier(self.plates_multiplier,
                                                  multiplier_parent))
                axes_msg = misc.axes_to_collapse(plates_m, parent.plates)
                m[i] = misc.sum_multiply(mask_i, m[i], r,
                                         axis=axes_msg,
                                         keepdims=True)

                # Remove leading singular plates if the parent does not have
                # those plate axes.
                m[i] = misc.squeeze_to_dim(m[i], len(shape_parent))

            return m_function
            raise NotImplementedError()

        # Compact the message to a proper shape
        for i in range(len(m)):

            # Empty messages are given as None. We can ignore those.
            if m[i] is not None:

                try:
                    r = self.broadcasting_multiplier(self.plates_multiplier,
                                                     multiplier_parent)
                except:
                    raise ValueError("The plate multipliers are incompatible. "
                                     "This node (%s) has %s and parent[%d] "
                                     "(%s) has %s"
                                     % (self.name,
                                        self.plates_multiplier,
                                        index,
                                        parent.name,
                                        multiplier_parent))

                ndim = len(parent.dims[i])
                # Source and target shapes
                if ndim > 0:
                    dims = misc.broadcasted_shape(np.shape(m[i])[-ndim:],
                                                  parent.dims[i])
                    from_shape = plates_self + dims
                else:
                    from_shape = plates_self
                to_shape = parent.get_shape(i)
                # Add variable axes to the mask
                mask_i = misc.add_trailing_axes(mask, ndim)
                # Apply mask and sum plate axes as necessary (and apply plate
                # multiplier)
                m[i] = r * misc.sum_multiply_to_plates(np.where(mask_i, m[i], 0),
                                                       to_plates=to_shape,
                                                       from_plates=from_shape,
                                                       ndim=0)

        return m

    def _message_from_children(self, u_self=None):
        msg = [np.zeros(shape) for shape in self.dims]
        #msg = [np.array(0.0) for i in range(len(self.dims))]
        isfunction = None
        for (child,index) in self.children:
            m = child._message_to_parent(index, u_parent=u_self)
            if callable(m):
                if isfunction is False:
                    raise NotImplementedError()
                elif isfunction is None:
                    msg = m
                else:
                    def join(m1, m2):
                        return (m1[0] + m2[0], m1[1] + m2[1])
                    msg = lambda x: join(m(x), msg(x))
                    isfunction = True
            else:
                if isfunction is True:
                    raise NotImplementedError()
                else:
                    isfunction = False
                    for i in range(len(self.dims)):
                        if m[i] is not None:
                            # Check broadcasting shapes
                            sh = misc.broadcasted_shape(self.get_shape(i), np.shape(m[i]))
                            try:
                                # Try exploiting broadcasting rules
                                msg[i] += m[i]
                            except ValueError:
                                msg[i] = msg[i] + m[i]

        return msg

    def _message_from_parents(self, exclude=None):
        return [list(parent._message_to_child())
                if ind != exclude else
                None
                for (ind,parent) in enumerate(self.parents)]

    def get_moments(self):
        raise NotImplementedError()

    def delete(self):
        """
        Delete this node and the children
        """
        for (ind, parent) in enumerate(self.parents):
            parent._remove_child(self, ind)
        for (child, _) in self.children:
            child.delete()

    @staticmethod
    def broadcasting_multiplier(plates, *args):
        return misc.broadcasting_multiplier(plates, *args)
        ## """
        ## Compute the plate multiplier for given shapes.

        ## The first shape is compared to all other shapes (using NumPy
        ## broadcasting rules). All the elements which are non-unit in the first
        ## shape but 1 in all other shapes are multiplied together.

        ## This method is used, for instance, for computing a correction factor for
        ## messages to parents: If this node has non-unit plates that are unit
        ## plates in the parent, those plates are summed. However, if the message
        ## has unit axis for that plate, it should be first broadcasted to the
        ## plates of this node and then summed to the plates of the parent. In
        ## order to avoid this broadcasting and summing, it is more efficient to
        ## just multiply by the correct factor. This method computes that
        ## factor. The first argument is the full plate shape of this node (with
        ## respect to the parent). The other arguments are the shape of the message
        ## array and the plates of the parent (with respect to this node).
        ## """

        ## # Check broadcasting of the shapes
        ## for arg in args:
        ##     misc.broadcasted_shape(plates, arg)

        ## # Check that each arg-plates are a subset of plates?
        ## for arg in args:
        ##     if not misc.is_shape_subset(arg, plates):
        ##         raise ValueError("The shapes in args are not a sub-shape of "
        ##                          "plates.")

        ## r = 1
        ## for j in range(-len(plates),0):
        ##     mult = True
        ##     for arg in args:
        ##         # if -j <= len(arg) and arg[j] != 1:
        ##         if not (-j > len(arg) or arg[j] == 1):
        ##             mult = False
        ##     if mult:
        ##         r *= plates[j]
        ## return r

    def move_plates(self, from_plate, to_plate):
        return _MovePlate(self,
                          from_plate,
                          to_plate,
                          name=self.name + ".move_plates")

    def add_plate_axis(self, to_plate):
        return AddPlateAxis(to_plate)(self,
                                      name=self.name+".add_plate_axis")

    def __getitem__(self, index):
        return Slice(self, index,
                     name=(self.name+".__getitem__"))

    def has_plotter(self):
        """
        Return True if the node has a plotter
        """
        return callable(self._plotter)

    def set_plotter(self, plotter):
        self._plotter = plotter

    def plot(self, fig=None, **kwargs):
        """
        Plot the node distribution using the plotter of the node

        Because the distributions are in general very difficult to plot, the
        user must specify some functions which performs the plotting as
        wanted. See, for instance, bayespy.plot.plotting for available plotters,
        that is, functions that perform plotting for a node.
        """
        if fig is None:
            import matplotlib.pyplot as plt
            fig = plt.gcf()
        if callable(self._plotter):
            ax = self._plotter(self, fig=fig, **kwargs)
            fig.suptitle('q(%s)' % self.name)
            return ax
        else:
            raise Exception("No plotter defined, can not plot")


    @staticmethod
    def _compute_message(*arrays, plates_from=(), plates_to=(), ndim=0):
        """
        A general function for computing messages by sum-multiply

        The function computes the product of the input arrays and then sums to
        the requested plates.
        """

        # Check that the plates broadcast properly
        if not misc.is_shape_subset(plates_to, plates_from):
            raise ValueError("plates_to must be broadcastable to plates_from")

        # Compute the explicit shape of the product
        shapes = [np.shape(array) for array in arrays]
        arrays_shape = misc.broadcasted_shape(*shapes)

        # Compute plates and dims that are present
        if ndim == 0:
            arrays_plates = arrays_shape
            dims = ()
        else:
            arrays_plates = arrays_shape[:-ndim]
            dims = arrays_shape[-ndim:]

        # Compute the correction term.  If some of the plates that should be
        # summed are actually broadcasted, one must multiply by the size of the
        # corresponding plate
        r = Node.broadcasting_multiplier(plates_from, arrays_plates, plates_to)

        # For simplicity, make the arrays equal ndim
        arrays = misc.make_equal_ndim(*arrays)

        # Keys for the input plates: (N-1, N-2, ..., 0)
        nplates = len(arrays_plates)
        in_plate_keys = list(range(nplates-1, -1, -1))

        # Keys for the output plates
        out_plate_keys = [key
                          for key in in_plate_keys
                          if key < len(plates_to) and plates_to[-key-1] != 1]

        # Keys for the dims
        dim_keys = list(range(nplates, nplates+ndim))

        # Total input and output keys
        in_keys = len(arrays) * [in_plate_keys + dim_keys]
        out_keys = out_plate_keys + dim_keys

        # Compute the sum-product with correction
        einsum_args = misc.zipper_merge(arrays, in_keys) + [out_keys]
        y = r * np.einsum(*einsum_args)

        # Reshape the result and apply correction
        nplates_result = min(len(plates_to), len(arrays_plates))
        if nplates_result == 0:
            plates_result = []
        else:
            plates_result = [min(plates_to[ind], arrays_plates[ind])
                             for ind in range(-nplates_result, 0)]

        y = np.reshape(y, plates_result + list(dims))

        return y


from .deterministic import Deterministic


def slicelen(s, length=None):
    if length is not None:
        s = slice(*(s.indices(length)))
    return max(0, misc.ceildiv(s.stop - s.start, s.step))

class Slice(Deterministic):

    """
    Basic slicing for plates.

    Slicing occurs when index is a slice object (constructed by start:stop:step
    notation inside of brackets), an integer, or a tuple of slice objects and
    integers.

    Currently, accept slices, newaxis, ellipsis and integers. For instance, does
    not accept lists/tuples to pick multiple indices of the same axis.

    Ellipsis expand to the number of : objects needed to make a selection tuple
    of the same length as x.ndim. Only the first ellipsis is expanded, any
    others are interpreted as :.

    Similar to:
    http://docs.scipy.org/doc/numpy/reference/arrays.indexing.html#basic-slicing
    """


    def __init__(self, X, slices, **kwargs):

        self._moments = X._moments
        self._parent_moments = (X._moments,)

        # Force a list
        if not isinstance(slices, tuple):
            slices = [slices]
        else:
            slices = list(slices)

        #
        # Expand Ellipsis
        #

        # Compute the number of required axes and how Ellipsis is expanded
        num_axis = 0
        ellipsis_index = None
        for (k, s) in enumerate(slices):

            if misc.is_scalar_integer(s) or isinstance(s, slice):
                num_axis += 1

            elif s is None:
                pass

            elif s is Ellipsis:
                # Index is an ellipsis, e.g., [...]

                if ellipsis_index is None:
                    # Expand ...
                    ellipsis_index = k
                else:
                    # Interpret ... as :
                    num_axis += 1
                    slices[k] = slice(None)

            else:
                raise TypeError("Invalid argument type: {0}".format(s.__class__))

        if num_axis > len(X.plates):
            raise IndexError("Too many indices")

        # The number of plates that were not given explicit slicing (either
        # Ellipsis was used or the number of slices was smaller than the number
        # of plate axes)
        expand_len = len(X.plates) - num_axis

        if ellipsis_index is not None:
            # Replace Ellipsis with correct number of :
            k = ellipsis_index
            del slices[k]
            slices = slices[:k] + [slice(None)] * expand_len + slices[k:]
        else:
            # Add trailing : so that each plate has explicit slicing
            slices = slices + [slice(None)] * expand_len

        #
        # Preprocess indexing:
        # - integer indices to non-negative values
        # - slice start/stop values to non-negative
        # - slice start/stop values based on the size of the plate
        #

        # Index for parent plates
        j = 0

        for (k, s) in enumerate(slices):

            if misc.is_scalar_integer(s):
                # Index is an integer, e.g., [3]

                if s < 0:
                    # Handle negative index
                    s += X.plates[j]
                if s < 0 or s >= X.plates[j]:
                    raise IndexError("Index out of range")
                # Store the preprocessed integer index
                slices[k] = s
                j += 1

            elif isinstance(s, slice):
                # Index is a slice, e.g., [2:6]

                # Normalize the slice
                s = slice(*(s.indices(X.plates[j])))
                if slicelen(s) <= 0:
                    raise IndexError("Slicing leads to empty plates")
                slices[k] = s
                j += 1

        self.slices = slices

        super().__init__(X,
                         dims=X.dims,
                         **kwargs)

    def _plates_to_parent(self, index):
        return self.parents[index].plates

    def _plates_from_parent(self, index):

        plates = list(self.parents[index].plates)

        # Compute the plates. Note that Ellipsis has already been preprocessed
        # to a proper number of :
        k = 0
        for s in self.slices:
            # Then, each case separately: slice, newaxis, integer

            if isinstance(s, slice):
                # Slice, e.g., [2:5]
                N = slicelen(s)
                if N <= 0:
                    raise IndexError("Slicing leads to empty plates")
                plates[k] = N
                k += 1

            elif s is None:
                # [np.newaxis]
                plates = plates[:k] + [1] + plates[k:]
                k += 1

            elif misc.is_scalar_integer(s):
                # Integer, e.g., [3]
                del plates[k]
            else:
                raise RuntimeError("BUG: Unknown index type. Should capture earlier.")

        return tuple(plates)

    @staticmethod
    def __reverse_indexing(slices, m_child, plates, dims):
        """
        A helpful function for performing reverse indexing/slicing
        """

        j = -1 # plate index for parent
        i = -1 # plate index for child
        child_slices = ()
        parent_slices = ()
        msg_plates = ()

        # Compute plate axes in the message from children
        ndim = len(dims)
        if ndim > 0:
            m_plates = np.shape(m_child)[:-ndim]
        else:
            m_plates = np.shape(m_child)

        for s in reversed(slices):

            if misc.is_scalar_integer(s):
                # Case: integer
                parent_slices = (s,) + parent_slices
                msg_plates = (plates[j],) + msg_plates
                j -= 1
            elif s is None:
                # Case: newaxis
                if -i <= len(m_plates):
                    child_slices = (0,) + child_slices
                i -= 1
            elif isinstance(s, slice):
                # Case: slice
                if -i <= len(m_plates):
                    child_slices = (slice(None),) + child_slices
                parent_slices = (s,) + parent_slices
                if ((-i > len(m_plates) or m_plates[i] == 1)
                    and slicelen(s) == plates[j]):
                    # Broadcasting can be applied. The message does not need
                    # to be explicitly shaped to the full size
                    msg_plates = (1,) + msg_plates
                else:
                    # No broadcasting. Must explicitly form the full size
                    # axis
                    msg_plates = (plates[j],) + msg_plates
                j -= 1
                i -= 1
            else:
                raise RuntimeError("BUG: Unknown index type. Should capture earlier.")

        # Set the elements of the message
        m_parent = np.zeros(msg_plates + dims)
        if np.ndim(m_parent) == 0 and np.ndim(m_child) == 0:
            m_parent = m_child
        elif np.ndim(m_parent) == 0:
            m_parent = m_child[child_slices]
        elif np.ndim(m_child) == 0:
            m_parent[parent_slices] = m_child
        else:
            m_parent[parent_slices] = m_child[child_slices]

        return m_parent


    def _compute_weights_to_parent(self, index, weights):
        """
        Compute the mask to the parent node.
        """
        if index != 0:
            raise ValueError("Invalid index")
        parent = self.parents[0]

        return self.__reverse_indexing(self.slices,
                                       weights,
                                       parent.plates,
                                       ())


    def _compute_message_to_parent(self, index, m, u):
        """
        Compute the message to a parent node.
        """

        if index != 0:
            raise ValueError("Invalid index")
        parent = self.parents[0]

        # Apply reverse indexing for the message arrays
        msg = [self.__reverse_indexing(self.slices,
                                       m_child,
                                       parent.plates,
                                       dims)
               for (m_child, dims) in zip(m, parent.dims)]

        return msg

    def _compute_moments(self, u):
        """
        Get the moments with an added plate axis.
        """

        # Process each moment
        for n in range(len(u)):

            # Compute the effective plates in the message/moment
            ndim = len(self.dims[n])
            if ndim > 0:
                shape = np.shape(u[n])[:-ndim]
            else:
                shape = np.shape(u[n])

            # Construct a list of slice objects
            u_slices = []

            # Index for the shape
            j = -len(self.parents[0].plates)

            for (k, s) in enumerate(self.slices):

                if s is None:
                    # [np.newaxis]
                    if -j < len(shape):
                        # Only add newaxis if there are some axes before
                        # this. It does not make any difference if you added
                        # leading unit axes
                        u_slices.append(s)

                else:
                    # slice or integer index

                    if -j <= len(shape):
                        # The moment has this axis, so it is not broadcasting it
                        if shape[j] != 1:
                            # Use the slice as it is
                            u_slices.append(s)
                        elif isinstance(s, slice):
                            # Slice.
                            # The moment is using broadcasting, just pick the
                            # first element but use slice in order to keep the
                            # axis
                            u_slices.append(slice(0,1,1))
                        else:
                            # Integer.
                            # The moment is using broadcasting, just pick the
                            # first element
                            u_slices.append(0)

                    j += 1

            # Slice the message/moment
            u[n] = u[n][tuple(u_slices)]

        return u

def AddPlateAxis(to_plate):

    if to_plate >= 0:
        raise Exception("Give negative value for axis index to_plate.")

    class _AddPlateAxis(Deterministic):

        def __init__(self, X, **kwargs):

            nonlocal to_plate

            N = len(X.plates) + 1

            # Check the parameters
            if to_plate >= 0 or to_plate < -N:
                raise ValueError("Invalid plate position to add.")

            # Use positive indexing only
            ## if to_plate < 0:
            ##     to_plate += N
            # Use negative indexing only
            if to_plate >= 0:
                to_plate -= N
                #self.to_plate = to_plate

            super().__init__(X,
                             dims=X.dims,
                             **kwargs)

        def _plates_to_parent(self, index):
            plates = list(self.plates)
            plates.pop(to_plate)
            return tuple(plates)
        #return self.plates[:to_plate] + self.plates[(to_plate+1):]

        def _plates_from_parent(self, index):
            plates = list(self.parents[index].plates)
            plates.insert(len(plates)-to_plate+1, 1)
            return tuple(plates)


        def _compute_weights_to_parent(self, index, weights):
            # Remove the added mask plate
            if abs(to_plate) <= np.ndim(weights):
                sh_weighs = list(np.shape(weights))
                sh_weights.pop(to_plate)
                weights = np.reshape(weights, sh_weights)
            return weights


        def _compute_message_to_parent(self, index, m, *u_parents):
            """
            Compute the message to a parent node.
            """

            # Remove the added message plate
            for i in range(len(m)):
                # Remove the axis
                if np.ndim(m[i]) >= abs(to_plate) + len(self.dims[i]):
                    axis = to_plate - len(self.dims[i])
                    sh_m = list(np.shape(m[i]))
                    sh_m.pop(axis)
                    m[i] = np.reshape(m[i], sh_m)

            return m

        def _compute_moments(self, u):
            """
            Get the moments with an added plate axis.
            """

            # Get parents' moments
            #u = self.parents[0].message_to_child()

            # Move a plate axis
            u = list(u)
            for i in range(len(u)):
                # Make sure the moments have all the axes
                #diff = len(self.plates) + len(self.dims[i]) - np.ndim(u[i]) - 1
                #u[i] = misc.add_leading_axes(u[i], diff)

                # The location of the new axis/plate:
                axis = np.ndim(u[i]) - abs(to_plate) - len(self.dims[i]) + 1
                if axis > 0:
                    # Add one axes to the correct position
                    sh_u = list(np.shape(u[i]))
                    sh_u.insert(axis, 1)
                    u[i] = np.reshape(u[i], sh_u)

            return u

    return _AddPlateAxis


class NodeConstantScalar(Node):
    @staticmethod
    def compute_fixed_u_and_f(x):
        """ Compute u(x) and f(x) for given x. """
        return ([x], 0)

    def __init__(self, a, **kwargs):
        self.u = [a]
        super().__init__(self,
                         plates=np.shape(a),
                         dims=[()],
                         **kwargs)

    def start_optimization(self):
        # FIXME: Set the plate sizes appropriately!!
        x0 = self.u[0]
        #self.gradient = np.zeros(np.shape(x0))
        def transform(x):
            # E.g., for positive scalars you could have exp here.
            self.gradient = np.zeros(np.shape(x0))
            self.u[0] = x
        def gradient():
            # This would need to apply the gradient of the
            # transformation to the computed gradient
            return self.gradient

        return (x0, transform, gradient)

    def add_to_gradient(self, d):
        self.gradient += d

    def message_to_child(self, gradient=False):
        if gradient:
            return (self.u, [ [np.ones(np.shape(self.u[0])),
                               #self.gradient] ])
                               self.add_to_gradient] ])
        else:
            return self.u


    def stop_optimization(self):
        #raise Exception("Not implemented for " + str(self.__class__))
        pass



################################################################################
# Copyright (C) 2015 Jaakko Luttinen
#
# This file is licensed under the MIT License.
################################################################################


"""
Unit tests for `gamma` module.
"""

import numpy as np

from scipy import special

from numpy import testing

from .. import gaussian
from bayespy.nodes import (Gaussian,
                           GaussianARD,
                           GaussianGamma,
                           Gamma,
                           Wishart)

from ...vmp import VB

from bayespy.utils import misc
from bayespy.utils import linalg
from bayespy.utils import random

from bayespy.utils.misc import TestCase


class TestGamma(TestCase):


    def test_lower_bound_contribution(self):

        a = 15
        b = 21
        y = 4
        x = Gamma(a, b)
        x.observe(y)
        testing.assert_allclose(
            x.lower_bound_contribution(),
            (
                a * np.log(b) +
                (a - 1) * np.log(y) -
                b * y -
                special.gammaln(a)
            )
        )

        # Just one latent node so we'll get exact marginal likelihood
        #
        # p(Y) = p(Y,X)/p(X|Y) = p(Y|X) * p(X) / p(X|Y)
        a = 2.3
        b = 4.1
        x = 1.9
        y = 4.8
        tau = Gamma(a, b)
        Y = GaussianARD(x, tau)
        Y.observe(y)
        mu = x
        nu = 2 * a
        s2 = b / a
        a_post = a + 0.5
        b_post = b + 0.5*(y - x)**2
        tau.update()
        testing.assert_allclose(
            [-b_post, a_post],
            tau.phi
        )
        testing.assert_allclose(
            Y.lower_bound_contribution() + tau.lower_bound_contribution(), # + tau.g,
            (
                special.gammaln((nu+1)/2)
                - special.gammaln(nu/2)
                - 0.5 * np.log(nu)
                - 0.5 * np.log(np.pi)
                - 0.5 * np.log(s2)
                - 0.5 * (nu + 1) * np.log(
                    1 + (y - mu)**2 / (nu * s2)
                )
            )
        )

        return


class TestGammaGradient(TestCase):
    """Numerically check Riemannian gradient of several nodes.

    Using VB-EM update equations will take a unit length step to the
    Riemannian gradient direction.  Thus, the change caused by a VB-EM
    update and the Riemannian gradient should be equal.
    """

    def test_riemannian_gradient(self):
        """Test Riemannian gradient of a Gamma node."""

        #
        # Without observations
        #

        # Construct model
        a = np.random.rand()
        b = np.random.rand()
        tau = Gamma(a, b)
        # Random initialization
        tau.initialize_from_parameters(np.random.rand(),
                                       np.random.rand())
        # Initial parameters
        phi0 = tau.phi
        # Gradient
        g = tau.get_riemannian_gradient()
        # Parameters after VB-EM update
        tau.update()
        phi1 = tau.phi
        # Check
        self.assertAllClose(g[0],
                            phi1[0] - phi0[0])
        self.assertAllClose(g[1],
                            phi1[1] - phi0[1])

        #
        # With observations
        #

        # Construct model
        a = np.random.rand()
        b = np.random.rand()
        tau = Gamma(a, b)
        mu = np.random.randn()
        Y = GaussianARD(mu, tau)
        Y.observe(np.random.randn())
        # Random initialization
        tau.initialize_from_parameters(np.random.rand(),
                                       np.random.rand())
        # Initial parameters
        phi0 = tau.phi
        # Gradient
        g = tau.get_riemannian_gradient()
        # Parameters after VB-EM update
        tau.update()
        phi1 = tau.phi
        # Check
        self.assertAllClose(g[0],
                            phi1[0] - phi0[0])
        self.assertAllClose(g[1],
                            phi1[1] - phi0[1])

        pass



    def test_gradient(self):
        """Test standard gradient of a Gamma node."""
        D = 3

        np.random.seed(42)

        #
        # Without observations
        #

        # Construct model
        a = np.random.rand(D)
        b = np.random.rand(D)
        tau = Gamma(a, b)
        Q = VB(tau)
        # Random initialization
        tau.initialize_from_parameters(np.random.rand(D),
                                       np.random.rand(D))
        # Initial parameters
        phi0 = tau.phi
        # Gradient
        rg = tau.get_riemannian_gradient()
        g = tau.get_gradient(rg)
        # Numerical gradient
        eps = 1e-8
        p0 = tau.get_parameters()
        l0 = Q.compute_lowerbound(ignore_masked=False)
        g_num = [np.zeros(D), np.zeros(D)]
        for i in range(D):
            e = np.zeros(D)
            e[i] = eps
            p1 = p0[0] + e
            tau.set_parameters([p1, p0[1]])
            l1 = Q.compute_lowerbound(ignore_masked=False)
            g_num[0][i] = (l1 - l0) / eps
        for i in range(D):
            e = np.zeros(D)
            e[i] = eps
            p1 = p0[1] + e
            tau.set_parameters([p0[0], p1])
            l1 = Q.compute_lowerbound(ignore_masked=False)
            g_num[1][i] = (l1 - l0) / eps

        # Check
        self.assertAllClose(g[0],
                            g_num[0])
        self.assertAllClose(g[1],
                            g_num[1])

        #
        # With observations
        #

        # Construct model
        a = np.random.rand(D)
        b = np.random.rand(D)
        tau = Gamma(a, b)
        mu = np.random.randn(D)
        Y = GaussianARD(mu, tau)
        Y.observe(np.random.randn(D))
        Q = VB(Y, tau)
        # Random initialization
        tau.initialize_from_parameters(np.random.rand(D),
                                       np.random.rand(D))
        # Initial parameters
        phi0 = tau.phi
        # Gradient
        rg = tau.get_riemannian_gradient()
        g = tau.get_gradient(rg)
        # Numerical gradient
        eps = 1e-8
        p0 = tau.get_parameters()
        l0 = Q.compute_lowerbound(ignore_masked=False)
        g_num = [np.zeros(D), np.zeros(D)]
        for i in range(D):
            e = np.zeros(D)
            e[i] = eps
            p1 = p0[0] + e
            tau.set_parameters([p1, p0[1]])
            l1 = Q.compute_lowerbound(ignore_masked=False)
            g_num[0][i] = (l1 - l0) / eps
        for i in range(D):
            e = np.zeros(D)
            e[i] = eps
            p1 = p0[1] + e
            tau.set_parameters([p0[0], p1])
            l1 = Q.compute_lowerbound(ignore_masked=False)
            g_num[1][i] = (l1 - l0) / eps

        # Check
        self.assertAllClose(g[0],
                            g_num[0])
        self.assertAllClose(g[1],
                            g_num[1])

        pass



################################################################################
# Copyright (C) 2011-2012 Jaakko Luttinen
#
# This file is licensed under the MIT License.
################################################################################


import itertools
import numpy as np
import scipy as sp
import scipy.linalg.decomp_cholesky as decomp
import scipy.linalg as linalg
import scipy.special as special
import time
import profile
import scipy.spatial.distance as distance

from .node import Node
from .stochastic import Stochastic
from bayespy.utils.misc import *

# Computes log probability density function of the Gaussian
# distribution
def gaussian_logpdf(y_invcov_y,
                    y_invcov_mu,
                    mu_invcov_mu,
                    logdetcov,
                    D):

    return (-0.5*D*np.log(2*np.pi)
            -0.5*logdetcov
            -0.5*y_invcov_y
            +y_invcov_mu
            -0.5*mu_invcov_mu)


# m prior mean function
# k prior covariance function
# x data inputs
# z processed data outputs (z = inv(Cov) * (y-m(x)))
# U data covariance Cholesky factor
def gp_posterior_moment_function(m, k, x, y, noise=None):

    # Prior
    mu = m(x)[0]
    K = k(x,x)[0]
    if noise != None:
        K += noise

    #print('hereiamagain')
    #print(K)

    # Compute posterior GP

    N = len(y)
    if N == 0:
        U = None
        z = None
    else:
        U = chol(K)
        z = chol_solve(U, y-mu)

    def get_moments(xh, covariance=1, mean=True):
        (kh,) = k(x, xh)

        # Function for computing posterior moments
        if mean:
            # Mean vector
            mh = m(xh)
            if z != None:
                mh += np.dot(kh.T, z)
        else:
            mh = None
        if covariance:
            if covariance == 1:
                # Variance vector
                khh = k(xh)
                if U != None:
                    khh -= np.einsum('i...,i...', kh, chol_solve(U, kh))
            elif covariance == 2:
                # Full covariance matrix
                khh = k(xh,xh)
                if U != None:
                    khh -= np.dot(kh.T, chol_solve(U,kh))
        else:
            khh = None

        return [mh, khh]

    return get_moments

# m prior mean function
# k prior covariance function
# x data inputs
# z processed data outputs (z = inv(Cov) * (y-m(x)))
# U data covariance Cholesky factor
## def gp_multi_posterior_moment_function(m, k, x, y, noise=None):

##     # Prior
##     mu = m(x)[0]
##     K = k(x,x)[0]
##     if noise != None:
##         K += noise

##     #print('hereiamagain')
##     #print(K)

##     # Compute posterior GP

##     N = len(y)
##     if N == 0:
##         U = None
##         z = None
##     else:
##         U = chol(K)
##         z = chol_solve(U, y-mu)

##     def get_moments(xh, covariance=1, mean=True):
##         (kh,) = k(x, xh)

##         # Function for computing posterior moments
##         if mean:
##             # Mean vector
##             mh = m(xh)
##             if z != None:
##                 mh += np.dot(kh.T, z)
##         else:
##             mh = None
##         if covariance:
##             if covariance == 1:
##                 # Variance vector
##                 khh = k(xh)
##                 if U != None:
##                     khh -= np.einsum('i...,i...', kh, chol_solve(U, kh))
##             elif covariance == 2:
##                 # Full covariance matrix
##                 khh = k(xh,xh)
##                 if U != None:
##                     khh -= np.dot(kh.T, chol_solve(U,kh))
##         else:
##             khh = None

##         return [mh, khh]

##     return get_moments

def gp_cov_se(D2, overwrite=False):
    if overwrite:
        K = D2
        K *= -0.5
        np.exp(K, out=K)
    else:
        K = np.exp(-0.5*D2)
    return K

def gp_cov_delta(N):
    return np.identity(N)


def squared_distance(x1, x2):
    # Reshape arrays to 2-D arrays
    sh1 = np.shape(x1)[:-1]
    sh2 = np.shape(x2)[:-1]
    d = np.shape(x1)[-1]
    x1 = np.reshape(x1, (-1,d))
    x2 = np.reshape(x2, (-1,d))
    # Compute squared Euclidean distance
    D2 = distance.cdist(x1, x2, metric='sqeuclidean')
    # Reshape the result
    D2 = np.reshape(D2, sh1 + sh2)
    return D2

# General rule for the parameters for covariance functions:
#
# (value, [ [dvalue1, ...], [dvalue2, ...], [dvalue3, ...], ...])
#
# For instance,
#
# k = covfunc_se((1.0, []), (15, [ [1,update_grad] ]))
# K = k((x1, [ [dx1,update_grad] ]), (x2, []))
#
# Plain values are converted as:
# value  ->  (value, [])

def gp_standardize_input(x):
    if np.ndim(x) == 0:
        x = add_trailing_axes(x, 2)
    elif np.ndim(x) == 1:
        x = add_trailing_axes(x, 1)
    return x

def gp_preprocess_inputs(*args):
    args = list(args)
    if len(args) < 1 or len(args) > 2:
        raise Exception("Number of inputs must be one or two")
    if len(args) == 2:
        if args[0] is args[1]:
            args[0] = gp_standardize_input(args[0])
            args[1] = args[0]
        else:
            args[1] = gp_standardize_input(args[1])
            args[0] = gp_standardize_input(args[0])
    else:
        args[0] = gp_standardize_input(args[0])

    return args

def covfunc_delta(theta, *inputs, gradient=False):

    amplitude = theta[0]

    if gradient:
        gradient_amplitude = gradient[0]
    else:
        gradient_amplitude = []

    inputs = gp_preprocess_inputs(*inputs)

    # Compute distance and covariance matrix
    if len(inputs) == 1:
        # Only variance vector asked
        x = inputs[0]
        K = np.ones(np.shape(x)[:-1]) * amplitude**2

    else:
        # Full covariance matrix asked
        x1 = inputs[0]
        x2 = inputs[1]
        # Number of inputs x1
        N1 = np.shape(x1)[-2]

        # x1 == x2?
        if x1 is x2:
            delta = True
            # Delta covariance
            K = gp_cov_delta(N1) * amplitude**2
        else:
            delta = False
            # Number of inputs x2
            N2 = np.shape(x2)[-2]
            # Zero covariance
            K = np.zeros((N1,N2))

    # Gradient w.r.t. amplitude
    if gradient:
        for ind in range(len(gradient_amplitude)):
            gradient_amplitude[ind] = K * (2 * gradient_amplitude[ind] / amplitude)

    if gradient:
        return (K, gradient)
    else:
        return K

def covfunc_se(theta, *inputs, gradient=False):

    amplitude = theta[0]
    lengthscale = theta[1]

    ## print('in se')
    ## print(amplitude)
    ## print(lengthscale)

    if gradient:
        gradient_amplitude = gradient[0]
        gradient_lengthscale = gradient[1]
    else:
        gradient_amplitude = []
        gradient_lengthscale = []

    inputs = gp_preprocess_inputs(*inputs)

    # Compute covariance matrix
    if len(inputs) == 1:
        x = inputs[0]
        # Compute variance vector
        K = np.ones(np.shape(x)[:-1])
        K *= amplitude**2
        # Compute gradient w.r.t. lengthscale
        for ind in range(len(gradient_lengthscale)):
            gradient_lengthscale[ind] = np.zeros(np.shape(x)[:-1])
    else:
        x1 = inputs[0] / (lengthscale)
        x2 = inputs[1] / (lengthscale)
        # Compute distance matrix
        K = squared_distance(x1, x2)
        # Compute gradient partly
        if gradient:
            for ind in range(len(gradient_lengthscale)):
                gradient_lengthscale[ind] = K * ((lengthscale**-1) * gradient_lengthscale[ind])
        # Compute covariance matrix
        gp_cov_se(K, overwrite=True)
        K *= amplitude**2
        # Compute gradient w.r.t. lengthscale
        if gradient:
            for ind in range(len(gradient_lengthscale)):
                gradient_lengthscale[ind] *= K

    # Gradient w.r.t. amplitude
    if gradient:
        for ind in range(len(gradient_amplitude)):
            gradient_amplitude[ind] = K * (2 * gradient_amplitude[ind] / amplitude)

    # Return values
    if gradient:
        return (K, gradient)
    else:
        return K


class NodeCovarianceFunction(Node):

    def __init__(self, covfunc, *args, **kwargs):
        self.covfunc = covfunc

        params = list(args)
        for i in range(len(args)):
            # Check constant parameters
            if is_numeric(args[i]):
                params[i] = NodeConstant([np.asanyarray(args[i])],
                                         dims=[np.shape(args[i])])
                # TODO: Parameters could be constant functions? :)

        Node.__init__(self, *params, dims=[(np.inf, np.inf)], **kwargs)

    def message_to_child(self, gradient=False):

        params = [parent.message_to_child(gradient=gradient) for parent in self.parents]
        return self.covariance_function(*params)

    def covariance_function(self, *params):
        params = list(params)
        gradient_params = list()
        for ind in range(len(params)):
            if isinstance(params[ind], tuple):
                gradient_params.append(params[ind][1])
                params[ind] = params[ind][0][0]
            else:
                gradient_params.append([])
                params[ind] = params[ind][0]

        def cov(*inputs, gradient=False):

            if gradient:
                grads = [[grad[0] for grad in gradient_params[ind]]
                         for ind in range(len(gradient_params))]

                (K, dK) = self.covfunc(params,
                                       *inputs,
                                       gradient=grads)

                for ind in range(len(dK)):
                    for (grad, dk) in zip(gradient_params[ind], dK[ind]):
                        grad[0] = dk

                K = [K]
                dK = []
                for grad in gradient_params:
                    dK += grad
                return (K, dK)

            else:
                K = self.covfunc(params,
                                 *inputs,
                                 gradient=False)
                return [K]

        return cov


class NodeCovarianceFunctionSum(NodeCovarianceFunction):
    def __init__(self, *args, **kwargs):
        NodeCovarianceFunction.__init__(self,
                                        None,
                                        *args,
                                        **kwargs)

    def covariance_function(self, *covfuncs):
        def cov(*inputs, gradient=False):
            K_sum = 0
            if gradient:
                dK_sum = list()
            for k in covfuncs:
                if gradient:
                    (K, dK) = k(*inputs, gradient=gradient)
                    dK_sum += dK
                else:
                    K = k(*inputs, gradient=gradient)
                K_sum += K[0]

            if gradient:
                return ([K_sum], dK_sum)
            else:
                return [K_sum]

        return cov


class NodeCovarianceFunctionDelta(NodeCovarianceFunction):
    def __init__(self, amplitude, **kwargs):
        NodeCovarianceFunction.__init__(self,
                                        covfunc_delta,
                                        amplitude,
                                        **kwargs)


class NodeCovarianceFunctionSquaredExponential(NodeCovarianceFunction):
    def __init__(self, amplitude, lengthscale, **kwargs):
        NodeCovarianceFunction.__init__(self,
                                        covfunc_se,
                                        amplitude,
                                        lengthscale,
                                        **kwargs)

class NodeMultiCovarianceFunction(NodeCovarianceFunction):

    def __init__(self, *args, **kwargs):
        NodeCovarianceFunction.__init__(self,
                                        None,
                                        *args,
                                        **kwargs)

    def covfunc(self, *covfuncs):
        def cov(*inputs, gradient=False):
            K_sum = 0
            if gradient:
                dK_sum = list()
            for k in covfuncs:
                if gradient:
                    (K, dK) = k(*inputs, gradient=gradient)
                    dK_sum += dK
                else:
                    K = k(*inputs, gradient=gradient)
                K_sum += K[0]


            if gradient:
                return ([K_sum], dK_sum)
            else:
                return [K_sum]

        return cov



class NodeConstantGaussianProcess(Node):
    def __init__(self, f, **kwargs):

        self.f = f
        Node.__init__(self, dims=[(np.inf,)], **kwargs)

    def message_to_child(self, gradient=False):

        # Wrapper
        def func(x, gradient=False):
            if gradient:
                return ([self.f(x)], [])
            else:
                return [self.f(x)]

        return func


# At least for now, simplify this GP node such that a GP is either
# observed or latent. If it is observed, it doesn't take messages from
# children, actually, it should not even have children!



#class NodeMultiGaussianProcess(NodeVariable):
class NodeMultiGaussianProcess(Stochastic):


    def __init__(self, m, k, **kwargs):

        self.x = []
        self.f = []

        # By default, posterior == prior
        self.m = m
        self.k = k

        # Ignore plates
        NodeVariable.__init__(self,
                              m,
                              k,
                              plates=(),
                              dims=[(np.inf,), (np.inf,np.inf)],
                              **kwargs)


    def message_to_parent(self, index):
        if index == 0:
            k = self.parents[1].message_to_child()[0]
            K = k(self.x, self.x)
            return [self.x,
                    self.mu,
                    K]
        if index == 1:
            raise Exception("not implemented yet")

    def message_to_child(self):
        if self.observed:
            raise Exception("Observable GP should not have children.")
        return self.u

    def get_parameters(self):
        return self.u

    def observe(self, x, f):

        if np.ndim(x) == 1:
            if np.shape(f) != np.shape(x):
                print(np.shape(f))
                print(np.shape(x))
                raise Exception("Number of inputs and function values do not match")
        elif np.shape(f) != np.shape(x)[:-1]:
            print(np.shape(f))
            print(np.shape(x))
            raise Exception("Number of inputs and function values do not match")

        self.observed = True
        self.x = x
        self.f = f
        ## self.x_obs = x
        ## self.f_obs = f

    # You might want:
    # - mean for x
    # - covariance (and mean) for x
    # - variance (and mean) for x
    # - i.e., mean and/or (co)variance for x
    # - covariance for x1 and x2



    def lower_bound_contribution(self, gradient=False):
        m = self.parents[0].message_to_child(gradient=gradient)
        k = self.parents[1].message_to_child(gradient=gradient)
        ## m = self.parents[0].message_to_child(gradient=gradient)[0]
        ## k = self.parents[1].message_to_child(gradient=gradient)[0]

        # Prior
        if gradient:
            (mu, dmus) = m(self.x, gradient=True)
            (K, dKs) = k(self.x, self.x, gradient=True)
        else:
            mu = m(self.x)
            K = k(self.x, self.x)
            dmus = []
            dKs = []

        mu = mu[0]
        K = K[0]

        # Log pdf
        if self.observed:
            # Vector of f-mu
            f0 = np.vstack([(f-m) for (f,m) in zip(self.f,mu)])
            # Full covariance matrix
            K_full = np.bmat(K)

            try:
                U = chol(K_full)
            except linalg.LinAlgError:
                print('non positive definite, return -inf')
                return -np.inf
            z = chol_solve(U, f0)
            #print(K)
            L = gaussian_logpdf(np.dot(f0, z),
                                0,
                                0,
                                logdet_chol(U),
                                np.size(self.f))

            for (dmu, func) in dmus:
                # Derivative w.r.t. mean vector
                d = -np.sum(z)
                # Send the derivative message
                func += d
                #func(d)

            for (dK, func) in dKs:
                # Compute derivative w.r.t. covariance matrix
                d = 0.5 * (np.dot(z, np.dot(dK, z))
                           - np.trace(chol_solve(U, dK)))
                # Send the derivative message
                #print('add gradient')
                #func += d
                func(d)

        else:
            raise Exception('Not implemented yet')

        return L

        ## Let f1 be observed and f2 latent function values.

        # Compute <log p(f1,f2|m,k)>

        #L = gaussian_logpdf(sum_product(np.outer(self.f,self.f) + self.Cov,


        # Compute <log q(f2)>




    def update(self):

        # Messages from parents
        m = self.parents[0].message_to_child()
        k = self.parents[1].message_to_child()
        ## m = self.parents[0].message_to_child()[0]
        ## k = self.parents[1].message_to_child()[0]

        if self.observed:

            # Observations of this node
            self.u = gp_posterior_moment_function(m, k, self.x, self.f)

        else:

            x = np.array([])
            y = np.array([])
            # Messages from children
            for (child,index) in self.children:
                (msg, mask) = child.message_to_parent(index)

                # Ignoring masks and plates..

                # m[0] is the inputs
                x = np.concatenate((x, msg[0]), axis=-2)

                # m[1] is the observations
                y = np.concatenate((y, msg[1]))

                # m[2] is the covariance matrix
                V = linalg.block_diag(V, msg[2])

            self.u = gp_posterior_moment_function(m, k, x, y, covariance=V)
            self.x = x
            self.f = y





class NodeGaussianProcess(Stochastic):
    #class NodeGaussianProcess(NodeVariable):

    def __init__(self, m, k, **kwargs):

        self.x = np.array([])
        self.f = np.array([])
        ## self.x_obs = np.zeros((0,1))
        ## self.f_obs = np.zeros((0,))

        # By default, posterior == prior
        self.m = m
        self.k = k

        # Ignore plates
        NodeVariable.__init__(self,
                              m,
                              k,
                              plates=(),
                              dims=[(np.inf,), (np.inf,np.inf)],
                              **kwargs)


    def message_to_parent(self, index):
        if index == 0:
            k = self.parents[1].message_to_child()[0]
            K = k(self.x, self.x)
            return [self.x,
                    self.mu,
                    K]
        if index == 1:
            raise Exception("not implemented yet")

    def message_to_child(self):
        if self.observed:
            raise Exception("Observable GP should not have children.")
        return self.u

    def get_parameters(self):
        return self.u

    def observe(self, x, f):

        if np.ndim(x) == 1:
            if np.shape(f) != np.shape(x):
                print(np.shape(f))
                print(np.shape(x))
                raise Exception("Number of inputs and function values do not match")
        elif np.shape(f) != np.shape(x)[:-1]:
            print(np.shape(f))
            print(np.shape(x))
            raise Exception("Number of inputs and function values do not match")

        self.observed = True
        self.x = x
        self.f = f
        ## self.x_obs = x
        ## self.f_obs = f

    # You might want:
    # - mean for x
    # - covariance (and mean) for x
    # - variance (and mean) for x
    # - i.e., mean and/or (co)variance for x
    # - covariance for x1 and x2



    def lower_bound_contribution(self, gradient=False):
        m = self.parents[0].message_to_child(gradient=gradient)
        k = self.parents[1].message_to_child(gradient=gradient)
        ## m = self.parents[0].message_to_child(gradient=gradient)[0]
        ## k = self.parents[1].message_to_child(gradient=gradient)[0]

        # Prior
        if gradient:
            (mu, dmus) = m(self.x, gradient=True)
            (K, dKs) = k(self.x, self.x, gradient=True)
        else:
            mu = m(self.x)
            K = k(self.x, self.x)
            dmus = []
            dKs = []

        mu = mu[0]
        K = K[0]

        # Log pdf
        if self.observed:
            f0 = self.f - mu

            #print('hereiam')
            #print(K)
            try:
                U = chol(K)
            except linalg.LinAlgError:
                print('non positive definite, return -inf')
                return -np.inf
            z = chol_solve(U, f0)
            #print(K)
            L = gaussian_logpdf(np.dot(f0, z),
                                0,
                                0,
                                logdet_chol(U),
                                np.size(self.f))

            for (dmu, func) in dmus:
                # Derivative w.r.t. mean vector
                d = -np.sum(z)
                # Send the derivative message
                func += d
                #func(d)

            for (dK, func) in dKs:
                # Compute derivative w.r.t. covariance matrix
                d = 0.5 * (np.dot(z, np.dot(dK, z))
                           - np.trace(chol_solve(U, dK)))
                # Send the derivative message
                #print('add gradient')
                #func += d
                func(d)

        else:
            raise Exception('Not implemented yet')

        return L

        ## Let f1 be observed and f2 latent function values.

        # Compute <log p(f1,f2|m,k)>

        #L = gaussian_logpdf(sum_product(np.outer(self.f,self.f) + self.Cov,


        # Compute <log q(f2)>




    def update(self):

        # Messages from parents
        m = self.parents[0].message_to_child()
        k = self.parents[1].message_to_child()
        ## m = self.parents[0].message_to_child()[0]
        ## k = self.parents[1].message_to_child()[0]

        if self.observed:

            # Observations of this node
            self.u = gp_posterior_moment_function(m, k, self.x, self.f)

        else:

            x = np.array([])
            y = np.array([])
            # Messages from children
            for (child,index) in self.children:
                (msg, mask) = child.message_to_parent(index)

                # Ignoring masks and plates..

                # m[0] is the inputs
                x = np.concatenate((x, msg[0]), axis=-2)

                # m[1] is the observations
                y = np.concatenate((y, msg[1]))

                # m[2] is the covariance matrix
                V = linalg.block_diag(V, msg[2])

            self.u = gp_posterior_moment_function(m, k, x, y, covariance=V)
            self.x = x
            self.f = y



################################################################################
# Copyright (C) 2013-2014 Jaakko Luttinen
#
# This file is licensed under the MIT License.
################################################################################


"""
Demonstrate linear Gaussian state-space model.

Some of the functions in this module are re-usable: 
  * ``model`` can be used to construct the classical linear state-space model.
  * ``infer`` can be used to apply linear state-space model to given data.
"""

import numpy as np
import scipy
import matplotlib.pyplot as plt

from bayespy.nodes import GaussianMarkovChain
from bayespy.nodes import Gaussian, GaussianARD
from bayespy.nodes import Gamma
from bayespy.nodes import SumMultiply
from bayespy.inference.vmp.nodes.gamma import diagonal

from bayespy.utils import random

from bayespy.inference.vmp.vmp import VB
from bayespy.inference.vmp import transformations

import bayespy.plot as bpplt


def model(M=10, N=100, D=3):
    """
    Construct linear state-space model.

    See, for instance, the following publication:
    "Fast variational Bayesian linear state-space model"
    Luttinen (ECML 2013)
    """

    # Dynamics matrix with ARD
    alpha = Gamma(1e-5,
                  1e-5,
                  plates=(D,),
                  name='alpha')
    A = GaussianARD(0,
                    alpha,
                    shape=(D,),
                    plates=(D,),
                    plotter=bpplt.GaussianHintonPlotter(rows=0, 
                                                        cols=1,
                                                        scale=0),
                    name='A')
    A.initialize_from_value(np.identity(D))

    # Latent states with dynamics
    X = GaussianMarkovChain(np.zeros(D),         # mean of x0
                            1e-3*np.identity(D), # prec of x0
                            A,                   # dynamics
                            np.ones(D),          # innovation
                            n=N,                 # time instances
                            plotter=bpplt.GaussianMarkovChainPlotter(scale=2),
                            name='X')
    X.initialize_from_value(np.random.randn(N,D))

    # Mixing matrix from latent space to observation space using ARD
    gamma = Gamma(1e-5,
                  1e-5,
                  plates=(D,),
                  name='gamma')
    gamma.initialize_from_value(1e-2*np.ones(D))
    C = GaussianARD(0,
                    gamma,
                    shape=(D,),
                    plates=(M,1),
                    plotter=bpplt.GaussianHintonPlotter(rows=0,
                                                        cols=2,
                                                        scale=0),
                    name='C')
    C.initialize_from_value(np.random.randn(M,1,D))

    # Observation noise
    tau = Gamma(1e-5,
                1e-5,
                name='tau')
    tau.initialize_from_value(1e2)

    # Underlying noiseless function
    F = SumMultiply('i,i', 
                    C, 
                    X,
                    name='F')
    
    # Noisy observations
    Y = GaussianARD(F,
                    tau,
                    name='Y')

    Q = VB(Y, F, C, gamma, X, A, alpha, tau, C)

    return Q


def infer(y, D, 
          mask=True, 
          maxiter=100,
          rotate=True,
          debug=False,
          precompute=False,
          update_hyper=0,
          start_rotating=0,
          plot_C=True,
          monitor=True,
          autosave=None):
    """
    Apply linear state-space model for the given data.
    """
        
    (M, N) = np.shape(y)

    # Construct the model
    Q = model(M, N, D)
    if not plot_C:
        Q['C'].set_plotter(None)
        
    if autosave is not None:
        Q.set_autosave(autosave, iterations=10)

    # Observe data
    Q['Y'].observe(y, mask=mask)

    # Set up rotation speed-up
    if rotate:
        
        # Initial rotate the D-dimensional state space (X, A, C)
        # Does not update hyperparameters
        rotA_init = transformations.RotateGaussianARD(Q['A'], 
                                                      axis=0,
                                                      precompute=precompute)
        rotX_init = transformations.RotateGaussianMarkovChain(Q['X'], 
                                                              rotA_init)
        rotC_init = transformations.RotateGaussianARD(Q['C'],
                                                      axis=0,
                                                      precompute=precompute)
        R_X_init = transformations.RotationOptimizer(rotX_init, rotC_init, D)

        # Rotate the D-dimensional state space (X, A, C)
        rotA = transformations.RotateGaussianARD(Q['A'], 
                                                 Q['alpha'],
                                                 axis=0,
                                                 precompute=precompute)
        rotX = transformations.RotateGaussianMarkovChain(Q['X'], 
                                                         rotA)
        rotC = transformations.RotateGaussianARD(Q['C'],
                                                 Q['gamma'],
                                                 axis=0,
                                                 precompute=precompute)
        R_X = transformations.RotationOptimizer(rotX, rotC, D)

        # Keyword arguments for the rotation
        if debug:
            rotate_kwargs = {'maxiter': 10,
                             'check_bound': True,
                             'check_gradient': True}
        else:
            rotate_kwargs = {'maxiter': 10}

    # Plot initial distributions
    if monitor:
        Q.plot()

    # Run inference using rotations
    for ind in range(maxiter):

        if ind < update_hyper:
            # It might be a good idea to learn the lower level nodes a bit
            # before starting to learn the upper level nodes.
            Q.update('X', 'C', 'A', 'tau', plot=monitor)
            if rotate and ind >= start_rotating:
                # Use the rotation which does not update alpha nor beta
                R_X_init.rotate(**rotate_kwargs)
        else:
            Q.update(plot=monitor)
            if rotate and ind >= start_rotating:
                # It might be a good idea to not rotate immediately because it
                # might lead to pruning out components too efficiently before
                # even estimating them roughly
                R_X.rotate(**rotate_kwargs)

    # Return the posterior approximation
    return Q


def simulate_data(M, N):
    """
    Generate a dataset using linear state-space model.

    The process has two latent oscillation components and one random walk
    component.
    """
    # Simulate some data
    D = 3
    c = np.random.randn(M, D)
    w = 0.3
    a = np.array([[np.cos(w), -np.sin(w), 0], 
                  [np.sin(w), np.cos(w),  0], 
                  [0,         0,          1]])
    x = np.empty((N,D))
    f = np.empty((M,N))
    y = np.empty((M,N))
    x[0] = 10*np.random.randn(D)
    f[:,0] = np.dot(c,x[0])
    y[:,0] = f[:,0] + 3*np.random.randn(M)
    for n in range(N-1):
        x[n+1] = np.dot(a,x[n]) + np.random.randn(D)
        f[:,n+1] = np.dot(c,x[n+1])
        y[:,n+1] = f[:,n+1] + 3*np.random.randn(M)

    return (y, f)


@bpplt.interactive
def demo(M=6, N=200, D=3, maxiter=100, debug=False, seed=42, rotate=True,
         precompute=False, plot=True, monitor=True):
    """
    Run the demo for linear state-space model.
    """

    # Use deterministic random numbers
    if seed is not None:
        np.random.seed(seed)

    # Get data
    (y, f) = simulate_data(M, N)

    # Add missing values randomly
    mask = random.mask(M, N, p=0.3)
    # Add missing values to a period of time
    mask[:,30:80] = False
    y[~mask] = np.nan # BayesPy doesn't require this. Just for plotting.

    # Run inference
    Q = infer(y, D,
              mask=mask,
              rotate=rotate,
              debug=debug,
              monitor=monitor,
              maxiter=maxiter)

    if plot:
        # Show results
        plt.figure()
        bpplt.timeseries_normal(Q['F'], scale=2)
        bpplt.timeseries(f, linestyle='-', color='b')
        bpplt.timeseries(y, linestyle='None', color='r', marker='.')


if __name__ == '__main__':

    import sys, getopt, os
    try:
        opts, args = getopt.getopt(sys.argv[1:],
                                   "",
                                   ["m=",
                                    "n=",
                                    "d=",
                                    "seed=",
                                    "maxiter=",
                                    "debug",
                                    "precompute",
                                    "no-plot",
                                    "no-monitor",
                                    "no-rotation"])
    except getopt.GetoptError:
        print('python lssm.py <options>')
        print('--m=<INT>        Dimensionality of data vectors')
        print('--n=<INT>        Number of data vectors')
        print('--d=<INT>        Dimensionality of the latent vectors in the model')
        print('--no-rotation    Do not apply speed-up rotations')
        print('--maxiter=<INT>  Maximum number of VB iterations')
        print('--seed=<INT>     Seed (integer) for the random number generator')
        print('--debug          Check that the rotations are implemented correctly')
        print('--no-plot        Do not plot the results')
        print('--no-monitor     Do not plot distributions during learning')
        print('--precompute     Precompute some moments when rotating. May '
              'speed up or slow down.')
        sys.exit(2)

    kwargs = {}
    for opt, arg in opts:
        if opt == "--no-rotation":
            kwargs["rotate"] = False
        elif opt == "--maxiter":
            kwargs["maxiter"] = int(arg)
        elif opt == "--debug":
            kwargs["debug"] = True
        elif opt == "--precompute":
            kwargs["precompute"] = True
        elif opt == "--seed":
            kwargs["seed"] = int(arg)
        elif opt in ("--m",):
            kwargs["M"] = int(arg)
        elif opt in ("--n",):
            kwargs["N"] = int(arg)
        elif opt in ("--d",):
            kwargs["D"] = int(arg)
        elif opt in ("--no-plot"):
            kwargs["plot"] = False
        elif opt in ("--no-monitor"):
            kwargs["monitor"] = False
        else:
            raise ValueError("Unhandled option given")

    demo(**kwargs)
    plt.show()



################################################################################
# Copyright (C) 2015 Jaakko Luttinen
#
# This file is licensed under the MIT License.
################################################################################


import numpy as np

from .node import Node, Moments
from .stochastic import Stochastic
from .deterministic import Deterministic


class DeltaMoments(Moments):
    r"""
    Class for the moments of constants or delta distributed variables
    """


    def compute_fixed_moments(self, x):
        r"""
        Compute the moments for a fixed value
        """
        return [x]


    def compute_dims_from_values(self, x):
        r"""
        Return the shape of the moments for a fixed value.
        """
        return ((),)


class DeltaToAny(Deterministic):
    r"""
    Special converter of delta moments to any moments
    """


    def __init__(self, X, moments):
        r"""
        """
        self._moments = moments
        self._parent_moments = [DeltaMoments()]
        #(plates, dims) = moments.compute_plates_and_dims(X.get_shape(0))
        dims = moments.compute_dims_from_shape(X.get_shape(0))
        super().__init__(X, dims=dims, **kwargs)
            

    def _compute_moments(self, u_X):
        r"""
        """
        x = u_X[0]
        return self._moments.compute_fixed_moments(x)
    

    def _compute_message_to_parent(self, index, m_child, u_X):
        r"""
        """
        # Convert child message array to a gradient function
        raise NotImplementedError()
        if index == 0:
            m = m_child[:2]
            return m
        else:
            raise ValueError("Invalid parent index")


    def _compute_weights_to_parent(self, index, weights):
        r"""
        """
        raise NotImplementedError()
        if index == 0:
            raise NotImplementedError()
        else:
            raise ValueError("Invalid parent index")


    def _plates_to_parent(self, index):
        r"""
        """
        
        raise NotImplementedError()
        if index == 0:
            self.get_shape(0)
            raise NotImplementedError()
        else:
            raise ValueError("Invalid parent index")


    def _plates_from_parent(self, index):
        r"""
        """
        raise NotImplementedError()
        if index == 0:
            return self.__cached_plates
            raise NotImplementedError()
        else:
            raise ValueError("Invalid parent index")



class Scalar(Stochastic):


    def __init__(self, plates=None):
        dims = [()]
        raise NotImplementedError()


    def get_riemannian_gradient(self):
        m_children = self._message_from_children()
        g = self.annealing * m_children[0]
        return g


    def get_gradient(self, rg):
        return rg


    def get_parameters(self):
        return self.u


    def set_parameters(self, x):
        if len(x) != 1:
            raise Exception("Wrong number of parameters. Should be 1, is {0}".format(len(x)))
        self.u = [x[0]]
        return


class PositiveScalar(Stochastic):
    pass


class Constant(Node):
    r"""
    Node for presenting constant values.

    The node wraps arrays into proper node type.
    """

    def __init__(self, moments, x, **kwargs):
        if not isinstance(moments, Moments) and issubclass(moments, Moments):
            raise ValueError("Give moments as an object instance instead of a class")
        self._moments = moments
        x = np.asanyarray(x)
        # Compute moments
        self.u = self._moments.compute_fixed_moments(x)
        # Dimensions of the moments
        dims = self._moments.compute_dims_from_values(x)
        # Resolve plates
        D = len(dims[0])
        if D > 0:
            plates = np.shape(self.u[0])[:-D]
        else:
            plates = np.shape(self.u[0])
        # Parent constructor
        super().__init__(dims=dims, plates=plates, **kwargs)


    def _get_id_list(self):
        """
        Returns the stochastic ID list.

        This method is used to check that same stochastic nodes are not direct
        parents of a node several times. It is only valid if there are
        intermediate stochastic nodes.

        To put it another way: each ID corresponds to one factor q(..) in the
        posterior approximation. Different IDs mean different factors, thus they
        mean independence. The parents must have independent factors.

        Stochastic nodes should return their unique ID. Deterministic nodes
        should return the IDs of their parents. Constant nodes should return
        empty list of IDs.
        """
        return []

    
    def get_moments(self):
        return self.u


    def set_value(self, x):
        x = np.asanyarray(x)
        shapes = [np.shape(ui) for ui in self.u]
        self.u = self._moments.compute_fixed_moments(x)
        for (i, shape) in enumerate(shapes):
            if np.shape(self.u[i]) != shape:
                raise ValueError("Incorrect shape for the array")


    def lower_bound_contribution(self, gradient=False, **kwargs):
        # Deterministic functions are delta distributions so the lower bound
        # contribuion is zero.
        return 0



################################################################################
# Copyright (C) 2014 Jaakko Luttinen
#
# This file is licensed under the MIT License.
################################################################################


"""
A module for the Bernoulli distribution node
"""

import numpy as np

from .binomial import (BinomialMoments,
                       BinomialDistribution)
from .expfamily import ExponentialFamily
from .beta import BetaMoments
from .node import Moments


class BernoulliMoments(BinomialMoments):
    """
    Class for the moments of Bernoulli variables.
    """


    def __init__(self):
        super().__init__(1)


class BernoulliDistribution(BinomialDistribution):
    """
    Class for the VMP formulas of Bernoulli variables.
    """


    def __init__(self):
        super().__init__(1)


class Bernoulli(ExponentialFamily):
    r"""
    Node for Bernoulli random variables.

    The node models a binary random variable :math:`z \in \{0,1\}` with prior
    probability :math:`p \in [0,1]` for value one:

    .. math::

        z \sim \mathrm{Bernoulli}(p).

    Parameters
    ----------
    p : beta-like node
        Probability of a successful trial

    Examples
    --------

    >>> import warnings
    >>> warnings.filterwarnings('ignore', category=RuntimeWarning)
    >>> from bayespy.nodes import Bernoulli, Beta
    >>> p = Beta([1e-3, 1e-3])
    >>> z = Bernoulli(p, plates=(10,))
    >>> z.observe([0, 1, 1, 1, 0, 1, 1, 1, 0, 1])
    >>> p.update()
    >>> import bayespy.plot as bpplt
    >>> import numpy as np
    >>> bpplt.pdf(p, np.linspace(0, 1, num=100))
    [<matplotlib.lines.Line2D object at 0x...>]
    """

    _moments = BernoulliMoments()
    _distribution = BernoulliDistribution()


    def __init__(self, p, **kwargs):
        """
        Create Bernoulli node.
        """
        super().__init__(p, **kwargs)


    @classmethod
    def _constructor(cls, p, **kwargs):
        """
        Constructs distribution and moments objects.
        """
        p = cls._ensure_moments(p, BetaMoments)
        parent_moments = (p._moments,)
        parents = [p]
        return ( parents,
                 kwargs,
                 ( (), ),
                 cls._total_plates(kwargs.get('plates'),
                                   cls._distribution.plates_from_parent(0, p.plates)),
                 cls._distribution,
                 cls._moments,
                 parent_moments)


    def __str__(self):
        """
        Print the distribution using standard parameterization.
        """
        p = 1 / (1 + np.exp(-self.phi[0]))
        return ("%s ~ Bernoulli(p)\n"
                "  p = \n"
                "%s\n"
                % (self.name, p))


from .deterministic import Deterministic
from .categorical import Categorical, CategoricalMoments

class CategoricalToBernoulli(Deterministic):
    """
    A node for converting 2-class categorical moments to Bernoulli moments.
    """

    
    def __init__(self, Z, **kwargs):
        """
        Create a categorical MC moments to categorical moments conversion node.
        """
        # Convert parent to proper type. Z must be a node.
        if not isinstance(Z._moments, CategoricalMoments):
            raise ValueError("Input node must be categorical")
        K = Z.dims[0][-1]
        if K != 2:
            raise Moments.NoConverterError("Only 2-class categorical can be converted to "
                                           "Bernoulli")
        dims = ( (), )
        self._moments = BernoulliMoments()
        self._parent_moments = (CategoricalMoments(2),)
        super().__init__(Z, dims=dims, **kwargs)

        
    def _compute_moments(self, u_Z):
        """
        Compute the moments given the moments of the parents.
        """
        u0 = u_Z[0][...,0]
        u = [u0]
        return u


    def _compute_message_to_parent(self, index, m, u_Z):
        """
        Compute the message to a parent.
        """
        if index == 0:
            m0 = np.concatenate([m[0][...,None],
                                 np.zeros(np.shape(m[0]))[...,None]],
                                axis=-1)
            return [m0]
        else:
            raise ValueError("Incorrect parent index")
    

# Make use of the conversion node
CategoricalMoments.add_converter(BernoulliMoments,
                                 CategoricalToBernoulli)



################################################################################
# Copyright (C) 2011-2012 Jaakko Luttinen
#
# This file is licensed under the MIT License.
################################################################################


import numpy as np
import matplotlib.pyplot as plt
import time

from bayespy.utils import misc
import bayespy.plot as bpplt
from bayespy.inference.vmp import nodes
from bayespy.inference.vmp.vmp import VB

def gaussianmix_model(N, K, D, covariance='full'):
    # N = number of data vectors
    # K = number of clusters
    # D = dimensionality
    
    # Construct the Gaussian mixture model

    # K prior weights (for components)
    alpha = nodes.Dirichlet(1e-3*np.ones(K),
                            name='alpha')
    # N K-dimensional cluster assignments (for data)
    z = nodes.Categorical(alpha,
                          plates=(N,),
                          name='z')
    if covariance.lower() == 'full':
        # K D-dimensional component means
        X = nodes.GaussianARD(0, 1e-3,
                            shape=(D,),
                            plates=(K,),
                            name='X')
        # K D-dimensional component covariances
        Lambda = nodes.Wishart(D, 0.01*np.identity(D),
                               plates=(K,),
                               name='Lambda')
        # N D-dimensional observation vectors
        Y = nodes.Mixture(z, nodes.Gaussian, X, Lambda, plates=(N,), name='Y')
    elif covariance.lower() == 'diagonal':
        # K D-dimensional component means
        X = nodes.GaussianARD(0, 1e-3,
                            plates=(D,K,),
                            name='X')
        # Inverse variances
        Lambda = nodes.Gamma(1e-3, 1e-3, plates=(D,K), name='Lambda')
        # N D-dimensional observation vectors
        Y = nodes.Mixture(z[...,None], nodes.GaussianARD, X, Lambda, plates=(N,D), name='Y')
    elif covariance.lower() == 'isotropic':
        # K D-dimensional component means
        X = nodes.GaussianARD(0, 1e-3,
                            plates=(D,K,),
                            name='X')
        # Inverse variances
        Lambda = nodes.Gamma(1e-3, 1e-3, plates=(D,K), name='Lambda')
        # N D-dimensional observation vectors
        Y = nodes.Mixture(z[...,None], nodes.GaussianARD, X, Lambda, plates=(N,D), name='Y')

    z.initialize_from_random()

    return VB(Y, X, Lambda, z, alpha)


@bpplt.interactive
def run(N=50, K=5, D=2):

    # Generate data
    N1 = int(np.floor(0.5*N))
    N2 = N - N1
    y = np.vstack([np.random.normal(0, 0.5, size=(N1,D)),
                   np.random.normal(10, 0.5, size=(N2,D))])

    
    # Construct model
    Q = gaussianmix_model(N,K,D)

    # Observe data
    Q['Y'].observe(y)

    # Run inference
    Q.update(repeat=30)

    # Run predictive model
    zh = nodes.Categorical(Q['alpha'], name='zh')
    Yh = nodes.Mixture(zh, nodes.Gaussian, Q['X'], Q['Lambda'], name='Yh')
    zh.update()

    # Plot predictive pdf
    N1 = 400
    N2 = 400
    x1 = np.linspace(-3, 15, N1)
    x2 = np.linspace(-3, 15, N2)
    xh = misc.grid(x1, x2)
    lpdf = Yh.integrated_logpdf_from_parents(xh, 0)
    pdf = np.reshape(np.exp(lpdf), (N2,N1))
    plt.clf()
    plt.contourf(x1, x2, pdf, 100)
    plt.scatter(y[:,0], y[:,1])
    print('integrated pdf:', np.sum(pdf)*(18*18)/(N1*N2))

    #Q['X'].show()
    #Q['alpha'].show()


if __name__ == '__main__':
    run()
    plt.show()

